/* 
    ColDet.cpp

	Collision detection for deformable models using kDOP trees

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "myvec.h"
#include "ColDet.h"
#include "Model.h"
#include <assert.h>

#pragma warning( disable : 4244 4305 )  

// dulezita konvence: prvni tri osy jsou x, y, z (jako AABB)
CDModel::CDModel(Model *m, int k, bool weapon)
{
	m_model = m;
	m_weapon = weapon;
	if (!weapon) m_numTri = m_model -> getNumTriangles();
	else m_numTri = m_model->m_pMeshes[m_model->m_numMeshes-1].m_numTriangles;
	// s modelem jsme kamaradi :-)
	m_pTri = new int[m_numTri];	
	root = NULL;
	assert(k%2 == 0);
	assert(k >= 6);
	m_k2 = k/2;
	m_axis[0][0] = 1; m_axis[0][1] = 0; m_axis[0][2] = 0;
	m_axis[1][0] = 0; m_axis[1][1] = 1; m_axis[1][2] = 0;	
	m_axis[2][0] = 0; m_axis[2][1] = 0; m_axis[2][2] = 1;
	switch (k) 
	{
		case 6:	break;
		case 14: // seriznuti rohu
			m_axis[3][0] = 1; m_axis[3][1] = 1; m_axis[3][2] = 1;
			m_axis[4][0] = 1; m_axis[4][1] =-1; m_axis[4][2] = 1;	
			m_axis[5][0] =-1; m_axis[5][1] = 1; m_axis[5][2] = 1;
			m_axis[6][0] = 1; m_axis[6][1] = 1; m_axis[6][2] =-1;
			break;
		case 26:
		case 18: // seriznuti hran
			m_axis[3][0] = 1; m_axis[3][1] = 1; m_axis[3][2] = 0;
			m_axis[4][0] = 1; m_axis[4][1] =-1; m_axis[4][2] = 0;	
			m_axis[5][0] = 1; m_axis[5][1] = 0; m_axis[5][2] = 1;
			m_axis[6][0] = 1; m_axis[6][1] = 0; m_axis[6][2] =-1;
			m_axis[7][0] = 0; m_axis[7][1] = 1; m_axis[7][2] = 1;
			m_axis[8][0] = 0; m_axis[8][1] = 1; m_axis[8][2] =-1;
			if (k==18) break;
		//case 26: seriznuti vrcholu i hran
			m_axis[9][0] = 1;  m_axis[9][1] = 1;  m_axis[9][2] = 1;
			m_axis[10][0] = 1; m_axis[10][1] =-1; m_axis[10][2] = 1;	
			m_axis[11][0] =-1; m_axis[11][1] = 1; m_axis[11][2] = 1;
			m_axis[12][0] = 1; m_axis[12][1] = 1; m_axis[12][2] =-1;
			break;
		default: assert(false); // spatne k
	}
}

CDModel::~CDModel()
{
	m_model = NULL;
	delete[] m_pTri;
	if (root) delTree(root);
}

// vraci true v pripade kolize
// t1 je vrchol meho stromu, t2 je vrchol stromu modelu *m_model2
bool CDModel::mycollide(DOPTree *t1, DOPTree *t2)
{
	if (t1==NULL || t2==NULL) return false; // prazdno nekoliduje ;-)
	if (!boxbox(t1, t2)) return false;
// obalove DOPy se protinaji:
	bool t1leaf = (t1->left == NULL) && (t1->right == NULL);
	bool t2leaf = (t2->left == NULL) && (t2->right == NULL);

	if (!t1leaf && !t2leaf)
	{ // oba vnitrni vrcholy
		float vol1 = (t1->max[0] - t1->min[0])*(t1->max[1] - t1->min[1])*(t1->max[2] - t1->min[2]);
		float vol2 = (t2->max[0] - t2->min[0])*(t2->max[1] - t2->min[1])*(t2->max[2] - t2->min[2]);
		if (vol1 > vol2) // delim prvni
			return mycollide(t1->left, t2) || mycollide(t1->right, t2);
		else // delim druhy
			return mycollide(t1, t2->left) || mycollide(t1, t2->right);
	}
	else if (t1leaf && !t2leaf)
	{
		return mycollide(t1, t2->left) || mycollide(t1, t2->right);
	}
	else if (!t1leaf && t2leaf)
	{
		return mycollide(t1->left, t2) || mycollide(t1->right, t2);
	}
	else // oba listy
	{
		for (int i=t1->imin; i<=t1->imax; i++)
			for (int j=t2->imin; j<=t2->imax; j++)
			{
				if (m_model->isNonhit(m_pTri[i]) || m_model2->m_model->isNonhit(m_model2->m_pTri[j]))
					continue; // aspon jeden z nich je nezasahovy trojuhelnik -> neni kolize
				if (tritri(m_pTri[i], m_model2->m_pTri[j])) 
				{
					//TRACE("Koliduji trky %i %i\n", m_pTri[i], m_model2->m_pTri[j]);
					//tritri(m_pTri[i], m_model2->m_pTri[j]);
					return true;
				}
			}
		return false;
	}
}

// vraci true v pripade kolize, navic nastavuje pole m_tr1 a m_tr2
// t1 je vrchol meho stromu, t2 je vrchol stromu modelu *m_model2
bool CDModel::mycollide2(DOPTree *t1, DOPTree *t2)
{
	if (t1==NULL || t2==NULL) return false; // prazdno nekoliduje ;-)
	if (!boxbox(t1, t2)) return false;
// obalove DOPy se protinaji:
	bool t1leaf = (t1->left == NULL) && (t1->right == NULL);
	bool t2leaf = (t2->left == NULL) && (t2->right == NULL);

	if (!t1leaf && !t2leaf)
	{ // oba vnitrni vrcholy
		float vol1 = (t1->max[0] - t1->min[0])*(t1->max[1] - t1->min[1])*(t1->max[2] - t1->min[2]);
		float vol2 = (t2->max[0] - t2->min[0])*(t2->max[1] - t2->min[1])*(t2->max[2] - t2->min[2]);
		if (vol1 > vol2) // delim prvni
		{
			// potrebuju vsechny kolidujici trky
			bool ret = mycollide2(t1->left, t2);
			ret = mycollide2(t1->right, t2)||ret;
			return ret;			
		}
		else // delim druhy
		{
			bool ret = mycollide2(t1, t2->left);
			ret = mycollide2(t1, t2->right)||ret;
			return ret;
		}
	}
	else if (t1leaf && !t2leaf)
	{		
		bool ret = mycollide2(t1, t2->left);
		ret = mycollide2(t1, t2->right)||ret;
		return ret;
	}
	else if (!t1leaf && t2leaf)
	{		
		bool ret = mycollide2(t1->left, t2);
		ret = mycollide2(t1->right, t2)||ret;
		return ret;		
	}
	else // oba listy
	{
		//TRACE("%i-DOP:  leaf(%i)-leaf(%i), celkem %i testiku\n", m_k2*2, (t1->imax - t1->imin + 1), (t2->imax - t2->imin + 1),
		//	(t1->imax - t1->imin + 1)*(t2->imax - t2->imin + 1));
		bool ret = false;
		for (int i=t1->imin; i<=t1->imax; i++)
			for (int j=t2->imin; j<=t2->imax; j++)
			{
				if (m_model->isNonhit(m_pTri[i]) || m_model2->m_model->isNonhit(m_model2->m_pTri[j]))
					continue; // aspon jeden z nich je nezasahovy trojuhelnik -> neni kolize
				if (tritri(m_pTri[i], m_model2->m_pTri[j])) 
				{
					//TRACE("Koliduji trky %i %i\n", m_pTri[i], m_model2->m_pTri[j]);					
					//if (m_tr1) m_tr1[m_pTri[i]] = true;
					//if (m_tr2) m_tr2[m_model2->m_pTri[j]] = true;
					m_model->setTriangleHit(m_pTri[i]);
					m_model2->m_model->setTriangleHit(m_model2->m_pTri[j]);
					ret = true;
				}
			}
		return ret;
	}
}

bool CDModel::collide(CDModel *m2)
{	
	assert(m2->m_k2 == m_k2);
	bbops = 0; ttops = 0;
	m_model2 = m2;	
	bool ret = mycollide(root, m2->root);
//	TRACE("Kolize: %i  boxbox testu: %i  tritri testu: %i\n", ret, bbops, ttops);
	return ret;

/* uplny, hloupy test:
		for (int i=0; i<m_numTri; i++)
			for (int j=0; j<m_model2->m_numTri; j++)
			{
				if (tritri(m_pTri[i], m_model2->m_pTri[j])) 
				{
					TRACE("Koliduji trky %i %i\n", m_pTri[i], m_model2->m_pTri[j]);
					tritri(m_pTri[i], m_model2->m_pTri[j]);
					return true;
				}
			}
		return false;
*/
}

bool CDModel::colTri(CDModel *m2)
{
	assert(m2->m_k2 == m_k2);
	bbops = 0; ttops = 0;
	m_model2 = m2;
	//m_tr1 = tr1; m_tr2 = tr2;	
	bool ret = mycollide2(root, m2->root);
	//TRACE("Kolize: %i  boxbox testu: %i  tritri testu: %i\n", ret, bbops, ttops);
	return ret;
}

void CDModel::delTree(DOPTree *v)
{
	if (v->left) delTree(v->left);
	if (v->right) delTree(v->right);
	delete v;
}

// vypocet obaloveho DOPu:
// Pozor, spravne by se skalarni soucin mel podelit normou vektoru m_axis,
// ale kdyz jen porovnavame, tak je to zbytecne
void CDModel::getMinMax(int zac, int kon, float min[MAX_DOP2], float max[MAX_DOP2])
{
	for (int n=0; n<m_k2; n++)  
	{
		min[n] = 1000000;
		max[n] = -1000000;
	}
	for (int i=zac; i<=kon; i++)
	{
		for (int k=0; k<3; k++)
		{
			float h[3];
			h[0] = m_model->m_pActVers[m_model -> m_pTriangles[m_pTri[i]].m_vertexIndices[k]].m_location[0];
			h[1] = m_model->m_pActVers[m_model -> m_pTriangles[m_pTri[i]].m_vertexIndices[k]].m_location[1];
			h[2] = m_model->m_pActVers[m_model -> m_pTriangles[m_pTri[i]].m_vertexIndices[k]].m_location[2];
			for (int n=0; n<m_k2; n++)
			{
				float dist = DOT(h, m_axis[n]);
				if (dist<min[n]) min[n] = dist;
				if (dist>max[n]) max[n] = dist;
			}
		}
	}
}

// je trojuhelnik negativni?
bool CDModel::negTri(int t, int osa, float delta)
{
	float stred;

	stred = (m_model->m_pActVers[m_model -> m_pTriangles[t].m_vertexIndices[0]].m_location[osa] +
			 m_model->m_pActVers[m_model -> m_pTriangles[t].m_vertexIndices[1]].m_location[osa] +
			 m_model->m_pActVers[m_model -> m_pTriangles[t].m_vertexIndices[2]].m_location[osa])/3;

/*
	if (osa==0)
	{
		stred = (m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[0]].x +
				 m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[1]].x +
				 m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[2]].x)/3;
	} else if (osa==1)
	{
		stred = (m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[0]].y +
				 m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[1]].y +
				 m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[2]].y)/3;
	} else
	{
		stred = (m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[0]].z +
				 m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[1]].z +
				 m_pVer[m_model -> m_pTriangles[t].m_vertexIndices[2]].z)/3;
	}
*/

//	TRACE("stred tr %i (osa %i) je %f\n",t, osa, stred);

	return stred<delta;
}

// prohazi trojuhelniky tak aby nejdriv byly zaporne (ty s tezistem na ose 'osa' < delta)
// vraci kde konci zaporne
int CDModel::prohazej(int zac, int kon, int osa, float delta)
{
	int i = zac; int j = kon;
	while (i<j)
	{
		while (negTri(m_pTri[i], osa, delta)&&i<j) i++;
		if (i==j)
		{
			if (negTri(m_pTri[i], osa, delta)) return i;
			else return i-1;
		}
		// i je 1. kladny
		while ((!negTri(m_pTri[j], osa, delta))&&i<j) j--;
		if (i==j) 
		{			
			return j-1;
		}
		// j je 1. zaporny; prohodime:
		int tmp = m_pTri[i];
		m_pTri[i] = m_pTri[j];
		m_pTri[j] = tmp;
		i++;
		j--;
	}
	if (i>j) return i-1;
	if (negTri(m_pTri[i], osa, delta)) return i;
	else return i-1;
}

// stavi strom nad trojuhelnicky s cisly m_pTri[zac], ..., m_pTri[kon]
DOPTree *CDModel::intBuild(int zac, int kon)
{
	DOPTree *ret = new DOPTree;
	ret->imin = zac; ret->imax = kon;
	ret->left = NULL; ret->right = NULL;
	getMinMax(zac, kon, ret->min, ret->max);
	float len[3]; int maxl = 0;
	len[0] = -1.0;
	// konvence ze prvni tri osy jsou x, y, z
	for (int k=0; k<3; k++) 
	{
		len[k] = ret->max[k] - ret->min[k];
		if (len[k] > len[maxl]) maxl = k;
	}
	if (zac==kon)
	{
		//TRACE("List s 1 trojuhelnikem\n");		
		return ret;
	}
	float delta = (ret->max[maxl] + ret->min[maxl]) / 2; // delim v pulce nejdelsi hrany
	int stred = prohazej(zac, kon, maxl, delta);
	if ((stred == kon || stred == zac-1)) // je-li list s vic nez jednim trojuhelnikem
	{
		maxl = (maxl+1)%3; // zkusim dalsi osu
		delta = (ret->max[maxl] + ret->min[maxl]) / 2; // delim v pulce nejdelsi hrany
	    stred = prohazej(zac, kon, maxl, delta);

		if ((stred == kon || stred == zac-1)) // je-li list s vic nez jednim trojuhelnikem
		{
			maxl = (maxl+1)%3; // zkusim dalsi osu
			delta = (ret->max[maxl] + ret->min[maxl]) / 2; // delim v pulce nejdelsi hrany
			stred = prohazej(zac, kon, maxl, delta);
		}
	}

// testik, jestli to vyslo spravne:	
	for (int n=zac; n<=stred; n++)
		if (!negTri(m_pTri[n], maxl, delta))
		{
			TRACE0("!\n"); 
			assert(false);
		}
	for (n=stred+1; n<=kon; n++)
	{
		if (negTri(m_pTri[n], maxl, delta))
		{
			TRACE0("!\n"); 
			assert(false);
		}	
	}

	if (stred == kon || stred == zac-1)
	{
		//TRACE("List s %i trojuhelniky\n", kon-zac+1);		
		return ret;
	}
	else // stred < kon
	{
		//TRACE("Deleni %i:%i\n", stred-zac+1, kon-stred);
		ret->left = intBuild(zac, stred);
	    ret->right = intBuild(stred+1, kon);	
	}
	return ret;
}

void CDModel::build()
{
	if (root) delTree(root);
	if (!m_weapon) for (int n=0; n<m_numTri; n++) m_pTri[n] = n;
	else  // jen posledni mesh:
		for (int n=0; n<m_numTri; n++) 
			m_pTri[n] = m_model->m_pMeshes[m_model->m_numMeshes-1].m_pTriangleIndices[n];

	root = intBuild(0, m_numTri-1);
	//TRACE("Hotovo ................................ \n\n");
}

void CDModel::myrefit(DOPTree *v)
{
	if ((v->left == NULL) && (v->right == NULL))
	{
		getMinMax(v->imin, v->imax, v->min, v->max);
	}
	else
	{ // neni list:
		for (int n=0; n<m_k2; n++) { v->min[n] = 1000000; v->max[n] = -1000000; }
		if (v->left) 
		{
			myrefit(v->left);
			for (int m=0; m<m_k2; m++) 
			{
				if (v->left->min[m] < v->min[m]) v->min[m]=v->left->min[m];
				if (v->left->max[m] > v->max[m]) v->max[m]=v->left->max[m];
			}
		}
		if (v->right) 
		{
			myrefit(v->right);
			for (int m=0; m<m_k2; m++) 
			{
				if (v->right->min[m] < v->min[m]) v->min[m]=v->right->min[m];
				if (v->right->max[m] > v->max[m]) v->max[m]=v->right->max[m];
			}
		}
	}
}

void CDModel::refit()
{
	myrefit(root);
}

// test jestli se dva intervaly protinaji
bool inline intTest(float a1, float b1, float a2, float b2)
{
	if (a1<a2) return b1>a2;
	else return b2>a1;
}

// test jestli se dva DOPy protinaji (z historickych duvodu nazev boxbox)
// Pozn.: muze vratit true (protinaji se) i kdyz jsou ve skutecnosti disjunktni
// (konzervativni (rychlejsi) test - viz. Klosowski)
bool CDModel::boxbox(DOPTree *t1, DOPTree *t2)
{	
	bbops++;
	for (int n=0; n<m_k2; n++)
	  if (!intTest(t1->min[n], t1->max[n], t2->min[n], t2->max[n])) return false;
	return true;
}

// test jestli se trojuhelniky protinaji
bool CDModel::tritri(int t1, int t2)
{
  int b0 = m_model->m_pTriangles[t1].m_vertexIndices[0];
  int b1 = m_model->m_pTriangles[t1].m_vertexIndices[1];
  int b2 = m_model->m_pTriangles[t1].m_vertexIndices[2];
  int c0 = m_model2->m_model->m_pTriangles[t2].m_vertexIndices[0];
  int c1 = m_model2->m_model->m_pTriangles[t2].m_vertexIndices[1];
  int c2 = m_model2->m_model->m_pTriangles[t2].m_vertexIndices[2];
/*
  float p0[3] = {m_pVer[b0].x, m_pVer[b0].y, m_pVer[b0].z};
  float p1[3] = {m_pVer[b1].x, m_pVer[b1].y, m_pVer[b1].z};
  float p2[3] = {m_pVer[b2].x, m_pVer[b2].y, m_pVer[b2].z};
  float q0[3] = {m_model2->m_pVer[c0].x, m_model2->m_pVer[c0].y, m_model2->m_pVer[c0].z};
  float q1[3] = {m_model2->m_pVer[c1].x, m_model2->m_pVer[c1].y, m_model2->m_pVer[c1].z};
  float q2[3] = {m_model2->m_pVer[c2].x, m_model2->m_pVer[c2].y, m_model2->m_pVer[c2].z};
*/
  ttops++;
  return 1==tri_contact(m_model->m_pActVers[b0].m_location, 
						m_model->m_pActVers[b1].m_location, 
						m_model->m_pActVers[b2].m_location, 
						m_model2->m_model->m_pActVers[c0].m_location,
						m_model2->m_model->m_pActVers[c1].m_location,
						m_model2->m_model->m_pActVers[c2].m_location);
}
	
void CDModel::getBBox(float min[3], float max[2])
{
	for (int n=0; n<3; n++)
	{
		min[n] = root->min[n];
		max[n] = root->max[n];
	}
}

void CDModel::draw(int h)
{
	mydraw(root, h, 0);
}

// 
void CDModel::mydraw(DOPTree *v, int h, int ah)
{
	if ((ah==h)||((v->left == NULL) && (v->right == NULL)))
	{
		float x0 = v->min[0], y0 = v->min[1], z0= v->min[2],
			  x1 = v->max[0], y1 = v->max[1], z1= v->max[2];
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);		
		glEnable(GL_BLEND);
		glBegin(GL_QUADS);
 			glColor4f(0.0, 1.0, 0.0, 0.7);
			glVertex3f(x0, y0, z0);
			glVertex3f(x1, y0, z0);
			glVertex3f(x1, y0, z1);
			glVertex3f(x0, y0, z1);
			glVertex3f(x0, y1, z0);
			glVertex3f(x1, y1, z0);
			glVertex3f(x1, y1, z1);
			glVertex3f(x0, y1, z1);		
			glVertex3f(x0, y0, z0);
			glVertex3f(x1, y0, z0);
			glVertex3f(x1, y1, z0);
			glVertex3f(x0, y1, z0);
			glVertex3f(x0, y0, z1);
			glVertex3f(x1, y0, z1);
			glVertex3f(x1, y1, z1);
			glVertex3f(x0, y1, z1);
			glVertex3f(x0, y0, z0);
			glVertex3f(x0, y0, z1);
			glVertex3f(x0, y1, z1);
			glVertex3f(x0, y1, z0);
			glVertex3f(x1, y0, z0);
			glVertex3f(x1, y0, z1);
			glVertex3f(x1, y1, z1);
			glVertex3f(x1, y1, z0);
		glEnd();
		glEnable(GL_CULL_FACE);
		glEnable(GL_LIGHTING);
		glDisable(GL_BLEND);
	} else {
		if (v->left) mydraw(v->left, h, ah+1);
		if (v->right) mydraw(v->right, h, ah+1);
	}
}