/* 
    Communication.h

	Routines for network communication using windows sockets.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#ifndef COMMUNICATION_H
#define COMMUNICATION_H

#include "LocalRemoteModel.h"
#include <afxmt.h>
#include <assert.h>

enum InfoMessages
{
	ERRMSG_NOMESSAGE, // zadna zprava jeste neprisla
	ERRMSG_TOOMANYCLIENTS, // spojeni odmitnuto: prilis mnoho klientu
	INITMSG_QUALITY0,
	INITMSG_QUALITY1
};
	
enum MessageTypes
{
	MSG_MODELSSTATE, 	
	MSG_INPUTEVENT,
	MSG_INFO,
	MSG_TRIANGLEHIT,
	MSG_JUDGEMSG,
	LAST_MESSAGE
};

const int cPosition = 1, cButton = 2, cKey = 3;	
typedef struct SInputEvent 
{	
	int player; // 0/1
	int type; // cPosition/cButton/cKey
	float x,y; // pro cPosition
	int data; // pro cButton a cKey
	bool flag; // pro cButton
} InputEvent;


const int cMaxJudgeMessage = 64;
typedef struct SJudgeMessage
{
	int	score0, score1; // je-li <0, tak nezobrazovat...
	char msg[cMaxJudgeMessage];

//	bool operator==(const struct SJudgeMessage &jm2)
//	{
//		if (score0==jm2.score0 && score1==jm2.score1 &&
//			strcmp(msg, jm2.msg) == 0) return true;
//		return false;
//	}
	SJudgeMessage() { msg[0]=0; }
} JudgeMessage;

// delky zprav v bajtech, poradi dano enumem MessageTypes
extern int msgLengths[LAST_MESSAGE];
	
//const cBUFLEN = (sizeof(ModelState)*2 > sizeof(InputEvent)) ? sizeof(ModelState)*2 : 
//				sizeof(InputEvent); // vybere maximum

const cBUFLEN = max(max(sizeof(ModelState)*2, sizeof(InputEvent)), sizeof(JudgeMessage));
	
typedef struct SMessage
{	
	char type; // z toho si odvodi delku
	char body[cBUFLEN];
} Message;	

const TYPE = 1;
const BODY = 2;

const cMaxClients = 16;

class CServerSocket;

// socket pro specialni paketovou komunikaci (17.2.03):
class CCommSocket : public CAsyncSocket
{
public:			
	CCommSocket(void *childView = 0, CServerSocket *ss = 0, int sn = 0);
	~CCommSocket();
	void reset();
	// nastavi m_modely, kterym budu posilat prisle zpravy o zasahu trojuhelniku:
	void setModels(Model *model1, Model *model2) 
		{ m_model[0] = model1; m_model[1] = model2; }
	// nastavi locModel cislo 'i' na 'lm' (by default 0):
	void setLocModel(int i, LocalModel *lm) 
		{ assert(i==0 || i==1); m_locModel[i] = lm; }
	// nastavi m_viewer (by default 0):
	void setViewer(void *v) { m_viewer = v; }

	bool isReady;	
	
	// prikonekti se k servru
	void startClient(const char *host, int port);
	// nastavi posledni kompletni stav sermire 'n' do promenne 'ms'; vraci false pokud zatim zadny nedorazil
	bool setLastModelState(int n, ModelState &ms);
	// posila novy stav sermiru:
	void sendModelState(const ModelState *ms[2]);
	// posila vstupni udalost:
	void sendInputEvent(const InputEvent &ie);
	// posila zpravu o zasazeni trojuhelnicku 'tri' hrace 'player':
	void sendTriangleHit(char player, int tri, byte value);
	// posila 1 bajtove info:
	void sendInfo(char err);
	// posila zpravu od rozhodciho:
	void sendJudgeMsg(int score0, int score1, const char *msg);

	int bi;
protected:		
	void *m_childView;
	Model *m_model[2]; // modely, kterym budu posilat prisle zpravy o zasahu trojuhelniku
	LocalModel *m_locModel[2]; // pro server: kam posilat InputEventy (je-li m_locModel[i] neNULL)
	void *m_viewer; // pro CModeViewer: kam posilat JudgeMsg

	// vnitrni fce na posilani paketu (vola se z sendXXX):
	void sendMessage(char typ, const char *p);
	
	int receiving;
	Message wrkMsg; // zprava, kterou prave prijimame (nemusi byt kompletni)
	ModelState m_modelState[2]; // v m_modelState jsou posledni kompletni (ne zpulky prepsane) stavy modelu; automaticky se prepisuji novyma	
	bool m_validModelState; // jestli uz dorazil aspon 1 modelState
	
	virtual void OnReceive( int nErrorCode );
	virtual void OnSend(int nErrorCode);
	virtual void OnClose(int nErrorCode);
	virtual void OnConnect(int nErrorCode);

	CMutex m_mutex;

	// pokud jsem na strane serveru:
	CServerSocket *m_serverSocket;
	int m_serverNr;	
	//// pripadna notifikace (nadrazene tridy) po prijeti paketu
	//virtual void OnPacket() { TRACE("Packet received\n"); }
};

// prijima/odmita dalsi klienty a udrzuje jejich CCommSockety:
class CServerSocket : public CAsyncSocket  
{
public:	
	bool m_newClient; // nastavi se pri acceptu noveho klienta; CChildView to pouzije pro preposlani vsech trojuhelniku a pak to smaze

	void killServer();

	CServerSocket(CFrameWnd *owner, int quality) 
	{		
		m_owner = owner; 
		assert(quality == 0 || quality == 1);
		m_quality = quality;
		m_newClient = false;
		for (int n=0; n<cMaxClients; n++) m_conn[n] = 0; 
	}
	~CServerSocket();
	void startServer(int port);
	void clientDied(int nr);
	// nastavi textovy popis i-teho klienta; vraci jestli je vubec nejaky pripojen:
	bool getClientString(int i, CString &txt) const; // spravne by to melo pouzivat mutex, aby se to netlouklo s OnAcceptem...
	// posila novy stav sermiru vsem pripojenym klientum:
	void sendAllModelState(const ModelState *ms[2])
	{ 
		for (int n=0; n<cMaxClients; n++) 
			if (m_conn[n]) m_conn[n]->sendModelState(ms);
	}
	void sendAllTriangleHit(char player, int tri, byte value)
	{
		for (int n=0; n<cMaxClients; n++) 
			if (m_conn[n]) m_conn[n]->sendTriangleHit(player, tri, value);
	}
	void sendAllJudgeMsg(int score0, int score1, const char *msg)
	{
		for (int n=0; n<cMaxClients; n++) 
			if (m_conn[n]) m_conn[n]->sendJudgeMsg(score0, score1, msg);
	}
	// rekne 'c'-temu komunikacnimu soketu, ze ma sve InputEventy pro hrace 'h' predavat
	// locModelu 'lm':
	void setLocModel(int c, int h, LocalModel *lm)
	{
		assert(0<=c && c<cMaxClients);
		assert(m_conn[c]);
		m_conn[c]->setLocModel(h, lm);
	}

protected:	
	int m_quality; // subdivided?
	virtual void OnAccept(int nErrorCode);
	CFrameWnd *m_owner;
	CCommSocket *m_conn[cMaxClients];		
};

#endif