/* 
    Ikan.cpp

	Inverse kinematics using IKAN system

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "Model.h"
#include "Ikan.h"
#include <assert.h>
#include <math.h>
#include "myvec.h"

#pragma warning( disable : 4244 4305 )  

// Deg -> Rad
inline void Deg2Rad(float &a) { a = a*M_PI/180; }

// Rad -> Deg
inline void Rad2Deg(float &a) { a = a*180/M_PI; }

// uhel da do <0, 2PI)
void normalizeAngle(float &a)
{
	while (a < 0) a += M_PI*2;
	while (a >= 2*M_PI) a -= M_PI*2;
}

// testuje jestli uhel x je mezi a a b
bool inRange(float a,float b, float x)
{
	normalizeAngle(a);
	normalizeAngle(b);
	normalizeAngle(x);
	if (a <= b) {
		if (a <= x && x <= b) return true;
		else return false;
	} else { /* a > b ... je to s prechodem pres 0 */
		if (a <= x || x <= b) return true;
		else return false;
	}
}

// vytahne z matice m swingovou a twistovou komponentu rotace (IKANske konvence)
void getSwingTwist(ikMatrix m, float &a, float &b, float &t)
{
  float sm[3], len;
  get_row(m, 2, sm); // osa z je smer paze
  len = norm(sm);
  sm[0]/=len; sm[1]/=len; sm[2]/=len;
  a = atan2( sm[0], sm[2] );
  b = -asin( sm[1] ); //24.7.02 pridano -
  ikMatrix ry, mul;
  rotation_principal_axis_to_ikMatrix('Y', -a, ry);
  rotation_principal_axis_to_ikMatrix('X', b, mul); //24.7.02 (odebrano -)
  hmatmult(ry, ry, mul);
  rotation_principal_axis_to_ikMatrix('Y', a, mul);
  hmatmult(ry, ry, mul);
// v ry je rotace o Beta kolem osy x orotovane o Alfa (kolem y)
/* -- spocita totez
  ikMatrix tst;
  float axis[3];
  get_row(mul, 0, axis);
  axisangletoikMatrix(ry, axis, b);
*/
  float yc[3], ycc[3];
  get_row(ry, 1, yc); // yc je osa y po provedeni swingu
  get_row(m, 1, ycc); // ycc je osa y po provedeni cele rotace, tj. swingu i twistu
  len = norm(yc); yc[0]/=len; yc[1]/=len; yc[2]/=len;
  len = norm(ycc); ycc[0]/=len; ycc[1]/=len; ycc[2]/=len;
  float d = DOT(yc, ycc);
  if (!( -1.001 <= d && d <= 1.001)) {
	  TRACE("Warning: d = %f\n",d);
  }
  t = acos(d);
}

void transposeIkMatrix(ikMatrix m)
{
 ikMatrix pom;
	for (int i=0;i<4;i++)
	  for (int j=0;j<4;j++)
	  {
		  pom[i][j] = m[j][i];
	  }
 	for (i=0;i<4;i++)
	  for (int j=0;j<4;j++)
	  {
		  m[i][j] = pom[i][j];
	  }
}

// oprava os udu; u je osa rotace udu
void getNormRotation(float u[3], ikMatrix mr, ikMatrix out)
{
	ikMatrix Sc; // pomocny souradny system
	float z[3] = {0, 0, 1}, len;
//	get_row(mr, 2, z);
//	len = norm(z); z[0]/=len; z[1]/=len; z[2]/=len;
	len = norm(u); u[0]/=len; u[1]/=len; u[2]/=len;
	float cosa = DOT(u, z),a;
	a = acos(cosa);
	float s[3] = {0, 0, 1}; //{(z[0]+u[0])/2, (z[1]+u[1])/2, (z[2]+u[2])/2};
	float pom[3] = {z[0]-u[0], z[1]-u[1], z[2]-u[2]}, osa[3];	
	len = norm(pom); pom[0]/=len; pom[1]/=len; pom[2]/=len;
	crossproduct(osa, pom, s);
	rotation_axis_to_ikMatrix(osa, a, Sc);
	// v Sc je ted novy souradny system (vzhledem ke staremu), jehoz osa z splyva s 'u'
	ikMatrix rSc, ScT;
	hmatmult(rSc, Sc, mr); // orotuju zadanou matici
	cpikMatrix(ScT, Sc); transposeIkMatrix(ScT);
	hmatmult(out, rSc, ScT);
}

// transoformacni matice z mych souradnic do IKANovskych:
ikMatrix ikan2my = {{0, 0, -1, 0}, {0, 1, 0, 0}, {1, 0, 0, 0}, {0, 0, 0, 1}};   //MT
// a zpatky:
ikMatrix my2ikan = {{0, 0, 1, 0}, {0, 1, 0, 0}, {-1, 0, 0, 0}, {0, 0, 0, 1}};	//M

// prevod z column-major do row-major a jeste transpozice
void Matrix2ikMatrix(const Matrix &m, ikMatrix im)
{
	for (int n = 0; n < 16; n++)
	{
		im[n/4][n%4] = m.getMatrix()[n];
	}
}

// prevod z row-major do column-major plus transpozice
void ikMatrix2Matrix(const ikMatrix im, Matrix &m)
{
  float p[16];
  for (int i=0;i<4;i++)
	  for (int j=0;j<4;j++)
	  {
		  p[i*4 + j] = im[i][j];
	  }
  m.set(p);
}

IKAN::IKAN(Model *m) {
	m_model = m;
// inicializuju SRSko
	m_shoulder = m->m_ikj[SHOULDER];
	m_elbow = m->m_ikj[ELBOW];
	m_hand = m->m_ikj[HAND];	
	
	// projekcni osa a kladny smer (uz v IKANovych souradnicich):
	float a[3] = {0, 1, 0}, p[3] = {1, 0, 0}; 	

	Matrix2ikMatrix(m->m_pJoints[m_elbow].m_relative, T); // u m_elbow je matice od ramene k lokti
	Matrix2ikMatrix(m->m_pJoints[m_hand].m_relative, S);

	// todle ma byt jako oprava na elbow to shoulder...
	//for (int i=0;i<3;i++) { T[3][i] *= -1; S[3][i] *= -1; }

	hmatmult(T, ikan2my, T);
	hmatmult(T, T, my2ikan);

	hmatmult(S, ikan2my, S);
	hmatmult(S, S, my2ikan);

	ikMatrix pom;
	hmatmult(pom, T, S);
	float dos[3];
	get_translation(pom, dos);
	m_dosah = norm(dos)*0.99; // 0.99 kvuli chybam

	srs.init(T, S, a, p);

	srs.ProjectOn();	

	// limity ramena
	amin[0] = -40; amax[0] = 130;
	bmin[0] = -80; bmax[0] = 80;
	tmax[0] = 60;
	// limity zapesti
	amin[1] = -50; amax[1] = 50;
	bmin[1] = -120; bmax[1] = 60; // nebo obracene?
	tmax[1] = 60;

	for (int i=0;i<2;i++)  {
	  Deg2Rad(amin[i]);Deg2Rad(amax[i]);
	  Deg2Rad(bmin[i]);Deg2Rad(bmax[i]);
	}
}

IKAN::~IKAN() {}

float IKAN::inEllipse(int i, float a, float b)
{
	float ca = (amin[i] + amax[i]) / 2, cb = (bmin[i] + bmax[i]) / 2;
	float alen = (amax[i] - amin[i]) / 2, blen = (bmax[i] - bmin[i]) / 2;
	float va = (a - ca)/alen;
	float vb = (b - cb)/blen;
	float f = va*va + vb*vb - 1;
	return f;	
}

// cim min, tim pohodlnejsi
// prvni a, b, t se tykaji ramene, dalsi zapesti
float IKAN::comfort(float a, float b, float t, float za, float zb, float zt)
{	
	za = za+zb+zt; //anti-warning
	return inEllipse(0, a, b) + t; // + (inEllipse(0, za, zb) + zt) / 4;
}

// predpokladam goal uz vzhledem k pozici ramene
void IKAN::scaleGoal( Vector &goal )
{
	if (goal.length() > m_dosah)  
	{		
		float scale = m_dosah/goal.length();
		goal.mul(scale, scale, scale);
		//TRACE("IKAN::scaleGoal - zmensuju %f -krat\n", scale);
	}
}

// prevadi pozici a orientaci ze swordholderu do ruky (presneji sh.final na hand.final)
void IKAN::SHGoal2HandGoal( Vector &goalPos, Matrix &orient)
{
/**/
	Matrix &mSwordHolder = m_model->m_pJoints[m_model->m_swordholder].m_relative;
	Matrix goal(orient);
	goal.setTranslation(goalPos.getVector());
	ikMatrix SHrel, invSHrel;
	Matrix2ikMatrix(mSwordHolder, SHrel);
	inverthomoikMatrix(invSHrel, SHrel);
	Matrix myinvSHrel;
	ikMatrix2Matrix(invSHrel, myinvSHrel);
	goal.postMultiply(myinvSHrel);

	float pom[3];
	goal.getTranslation(pom);
	goalPos.set(pom);
	float nul[3]={0,0,0};
	orient.set(goal.getMatrix());
	orient.setTranslation(nul);
/**/

/* old:
	float gp[3];
	goalPos.get(gp);
	Matrix &mSwordHolder = m_model->m_pJoints[m_model->m_swordholder].m_relative;
	mSwordHolder.inverseTranslateVect(gp);
	Matrix ir;
	mSwordHolder.inverseRotateMatrix(ir);
	orient.postMultiply(ir);
	goalPos.set(gp);
/**/
}

// nastavi cilovou pozici a orientaci EE (odpovida swordholder.final);
// vstupem je pozice a orientace swordholderu ve svetovych souradnicich a pozice loktu (v souradnicich rootu)
int IKAN::setGoal( Vector goalPos, Matrix orient, float elPos[3])
{
	return pomSetGoal(goalPos, orient, elPos, 2);
}

int IKAN::compElbowPos( Vector goalPos, Matrix orient, float elPos[3])
{
	return pomSetGoal(goalPos, orient, elPos, 1);
}

// spolecna rutina pro setGoal a compElbowPos
// mod 1: spocita elPos pomoci optimalizace pohodlnosti; nenastavuje rotace kloubu
// mod 2: spocita IK podle daneho elPos
int IKAN::pomSetGoal( Vector goalPos, Matrix orient, float elPos[3], int mode)
{
	Vector e;	
	Matrix &mShoulder = m_model->m_pJoints[m_shoulder].m_transform;
	Matrix &mElbow = m_model->m_pJoints[m_elbow].m_transform;
	Matrix &mHand = m_model->m_pJoints[m_hand].m_transform;


	// nejdriv prevedu na hand.final:
	SHGoal2HandGoal( goalPos, orient);


// aby to fungovalo, musi byt spocitane m_finaly !!!
	e.transform( m_model->m_pJoints[m_shoulder].m_final ); // v e je ted pozice ramena
    goalPos.sub(e); 	


	float ea[3];
	scaleGoal(goalPos);
	goalPos.get(ea);
	int m_neck = m_model->m_pJoints[m_shoulder].m_parent;
	m_model->m_pJoints[m_neck].m_final.inverseRotateVect(ea);

	ikMatrix G;

		ikMatrix transf;
		Matrix2ikMatrix(m_model->m_pJoints[m_neck].m_final, transf); // tim se transponuje
		Matrix2ikMatrix(orient, G);
		transposeIkMatrix(G);
		hmatmult(G, transf, G);
		transposeIkMatrix(G); // pro prevod predpokladam uz G transponovane

	set_translation(G, ea);
	hmatmult(G, ikan2my, G);
	hmatmult(G, G, my2ikan); // prevod do IKANovskych souradnic

	float rangle;
	int reachable = srs.SetGoal(G, rangle);	
	if (!reachable)  {
		TRACE("Unreachable\n");
		return 0;	
	}
	
	static float swivelAngle = 3.14;		
	ikMatrix shRot, wrRot; // rotace ramene	a zapesti
	if (mode==1)
	{ // automaticky vyber polohy lokte maximalizaci 'pohodlnosti':				
		//bool set = false;
		float swivelStart = 0;
		float bestSwiv = -1, mincomf = 10000;
		for (int n=0;n<360;n++)  
		{
			float sA;
			sA = swivelStart + n*M_PI/180;
			normalizeAngle(sA);
			srs.SolveR1R2(sA, shRot, wrRot);	  
			
			//float nadlokti[3];
			//ikMatrix normr;
			//get_translation(T, nadlokti);	
			//getNormRotation(nadlokti, shRot, normr); // cast se da predpocitat
			float a, b, t, za, zb, zt;
			getSwingTwist(shRot, a, b, t);
			getSwingTwist(wrRot, za, zb, zt);			
			
			float comf;
			if ((comf=comfort(a, b, t, za, zb, zt)) < mincomf)  
			{
				mincomf=comf; bestSwiv=sA;
			}
		}
		swivelAngle = bestSwiv;			
		srs.AngleToPos(swivelAngle, elPos);
		return 1;
	}
	else // mode == 2
	{ // mam recenou pozici lokte:
	//float pos[3];
	//srs.AngleToPos(swivelAngle, pos);
	//TRACE("pos = [%f, %f, %f]\n",pos[0], pos[1], pos[2]);
	//TRACE("pos = [%f, %f, %f]\n",elbowPos[0], elbowPos[1], elbowPos[2]);

		swivelAngle = srs.PosToAngle(elPos);
	}

	swivdeg = (swivelAngle*180)/M_PI;	
	//TRACE("swivdeg = %i\n", swivdeg);
	
	//if (!set) return 0;	
	srs.SolveR1R2(swivelAngle, shRot, wrRot);
	/*{
		float a, b, t, za, zb, zt;
		getSwingTwist(shRot, a, b, t);
		getSwingTwist(wrRot, za, zb, zt);
		TRACE("Shoulder: Swing = %f, %f, Twist = %f", a*180/M_PI, b*180/M_PI, t*180/M_PI);
		TRACE("inEll = %f, t = %f\n", inEllipse(0, a, b), t);
	}*/




	float a, b, t;

	float nadlokti[3];
	ikMatrix normr;
	get_translation(T, nadlokti);
	getNormRotation(nadlokti, shRot, normr);
	getSwingTwist(normr, a, b, t);
//	TRACE("Normalized shoulder: Swing = %f, %f, Twist = %f\n", a*180/M_PI, b*180/M_PI, t*180/M_PI);
	getSwingTwist(shRot, a, b, t);
//	TRACE("Shoulder: Swing = %f, %f, Twist = %f\n", a*180/M_PI, b*180/M_PI, t*180/M_PI);

	  float osaruky[3];
	  ikMatrix prostr, normw;
	  Matrix2ikMatrix(m_model->m_pJoints[m_model->m_midfinger].m_relative, prostr);
	  hmatmult(prostr, ikan2my, prostr);
	  hmatmult(prostr, prostr, my2ikan); // do IKANovskych souradnic
	  get_translation(prostr, osaruky);
	  getNormRotation(osaruky, wrRot, normw);
	  getSwingTwist(normw, a, b, t);
//	  TRACE("Normalized wrist: Swing = %f, %f, Twist = %f\n", a*180/M_PI, b*180/M_PI, t*180/M_PI);    
	  getSwingTwist(wrRot, a, b, t);
//	  TRACE("Wrist: Swing = %f, %f, Twist = %f\n", a*180/M_PI, b*180/M_PI, t*180/M_PI);    

	Matrix myShRot, My2ikan, Ikan2my;
	ikMatrix2Matrix(shRot, myShRot);
	ikMatrix2Matrix(my2ikan, My2ikan);
	ikMatrix2Matrix(ikan2my, Ikan2my);			
	/* -- zajima me jen zapesti */
	mShoulder.set(Ikan2my.getMatrix());	
	mShoulder.postMultiply(myShRot);
	mShoulder.postMultiply(My2ikan);	

	ikMatrix elRot; // rotace lokte
	Matrix myElRot;
	float osal[3] = {0, 1, 0};
	//static float angl = 0;
	axisangletoikMatrix(elRot, osal, rangle);

	ikMatrix2Matrix(elRot, myElRot);
	mElbow.set(Ikan2my.getMatrix());	
	mElbow.postMultiply(myElRot);
	mElbow.postMultiply(My2ikan);
	/**/

	Matrix myHaRot;
	ikMatrix2Matrix(wrRot, myHaRot);
	mHand.set(Ikan2my.getMatrix());	
	mHand.postMultiply(myHaRot);
	mHand.postMultiply(My2ikan);
	
	return 1;
}