/* 
    InputDevice.h

	Interace for input devices

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#ifndef INPUTDEVICE_H
#define INPUTDEVICE_H

#include "Vector.h"
#include "Matrix.h"
#include "myvec.h"
#include "Model.h"
#include "Win32Timer.h"
#include "KeyframeInterpolator.h"
#include "Judge.h"
#include <afxmt.h>

typedef struct SPoint2 {
	double x,y;
} Point2;	  

// trida na zjistovani pozice mysi:
class CMouseInputDevice {
protected:
	CRect m_rect; // vyrez v okne odpovidajici rozsahu pohybu mysi	
public:
	CMouseInputDevice(CRect rect) { m_rect = rect; }
	virtual ~CMouseInputDevice() {}
	// vrati aktualni normalizovanou polohu mysi (v (-1, 1))
	void getMousePos(float &x, float &y);
};

// trida (rozhrani) co definuje funkce informujici o zmene polohy a tlacitek
class C2DInputDevice  {
protected:
	float m_mx, m_my; // pozice v intervalu (-1, 1)
	float m_maxSpeed; // v jednotkach prostoru (-1,-1,1,1) za sekundu
	Win32Timer m_timer;

public:
	C2DInputDevice() { m_mx=0; m_my=0; m_maxSpeed = 4.0; m_timer.reset(); }
	virtual ~C2DInputDevice() {}	
	// vlozi novou pozici vstupniho zarizeni - nastavi m_mx, m_my podle x,y a maxSpeedu:
	virtual void feedNewPos(float x, float y);
	// zpravi tridu o kliknuti/pusteni tlacitka (WM_LBUTTONDOWN, WM_...)
	virtual void feedButton(int a) { a=a; }
};

/* Sphere1MID: Analogie sferickeho interpolatoru ve VRML
*/
/*
class Sphere1MID : public InputDevice  {
protected:
	float m_v; // polomer nejvetsi kruznice (v >= 1)		
public:
	Sphere1MID(Model *m) : MouseInputDevice(m) {}
	virtual ~Sphere1MID() {}
	// vraci matici orientace a pozice koncoveho efektoru
	virtual void getGoal(Matrix &or, Vector &opos);

	void setPars(CRect rect, float v, float al) {
		MouseInputDevice::setPars(rect, al);
		m_v = v;
	}

protected:
	// promitne do vnitrniho (jednotkoveho) kruhu
	// inner = false, pokud padlo nejriv mimo nej (tj. je to na druhe polokouli)
	void toCircle(float &icx, float &icy, bool &inner);
};
*/

/* KeyframeMID: Vypocet polohy a orientace na zaklade klicovych snimku
*/

class KeyframeMID : public C2DInputDevice  {	
protected:
	Model *m_model;
	int m_gridX, m_gridY; // velikost mrizky (m_gridX*m_gridY je pocet klic. poloh)
	Matrix *m_keyframes; // klicove polohy naprahu	
	Point3 *m_elbowPos; // klicove polohy ramena
public:
	KeyframeMID(Model *m, int gridX = 3, int gridY = 3);
	virtual ~KeyframeMID();
	// vraci matici orientace a pozice koncoveho efektoru
	virtual void getGoal(Matrix &or, Vector &opos);
	// totez co getGoal ale v IKANovske konvenci + navic jeste elbow position
	void getikMatrixEl(ikMatrix m, float elpos[3]);	

	// nahraje klicovou polohu na misto (x, y) v mrizce
	bool loadKeyframe(const char *fname, int x, int y, double sx=0, double sy=0, double sz=0);
	// nahraje klicove polohy lokte (v souradnicich ramene)
	bool loadElbowPos(const char *fname);

	// ulozi/nahraje vsechny klicove polohy naprahu do/ze souboru
	bool saveAllKeyframes(const char *fname);
	bool loadAllKeyframes(const char *fname);
};

/* ActionMID: Vstupni "zarizeni" podporujici naprahy, seky a kryty; vyuziva KeyframeMID
*/

// stavy:
const cNaprah = 1, cKryt = 2, cSek = 3, cNaprahKryt = 4, cNaprahSek = 5, 
	  cSekNaprah = 6, cKrytNaprah = 7, 
	  cSNK = 8, cSNKN = 9; // nouzove kryti, blize viz poznamky 18.7.02

// specialni zony (3.3.03):
const cZonaMimo = 1, cZonaNaprah = 2, cZonaBod = 3;

class ActionMID : public C2DInputDevice  {
protected:
	CMutex m_mutex; // na vzajemne vylouceni vice threadu (feedXXX volan taky z CCommSocket.OnReceive)
	Model *m_model;
	int m_stav;
	Win32Timer m_timer;
	double m_timerAdd, m_timerMul;
	double m_dobaNaprahKryt, m_dobaNaprahSek, m_dobaSNK, m_dobaSNKN, m_maxDobaSeku;
	ikMatrix m_from; // pro cSNK, cSNKN: okduk interpolovat
	float m_pfrom[3]; // totez pro pozici lokte
	// otestuje jestli cas vyprsel a kdyz ano tak zmeni stav
	void testTime();
	// z matice "ikRes" extrahuje rotaci "or" a posunuti "opos"
	void setPosOr(ikMatrix ikRes, Matrix &or, Vector &opos);
	// zjisti v jake zone se vstupni zarizeni (m_mx, m_my) prave ted nachazi:
	int getZone();

public:
	KeyframeMID m_naprahy, m_seky, m_kryty;
	ActionMID(Model *m, int nsx, int nsy, int ssx, int ssy, int ksx, int ksy); // velikosti mrizek pro polohy naprahu, seku a krytu
	virtual ~ActionMID() {}	
	// vraci pevnost drzeni zbrani od 0 do 1:
	float getToughness();
	// vraci matici orientace a pozice koncoveho efektoru
	virtual void getGoal(Matrix &or, Vector &opos);
	// totez co getGoal ale v IKANovske konvenci, navic pocita jeste elbow position
	void getikMatrixEl(ikMatrix m, float elpos[3]);
	// vraci interpolovanou pozici lokte
	virtual void getElbow(float elpos[3]);
	// informace o kliknuti/pusteni tlacitka (WM_LBUTTONDOWN, WM_...)
	virtual void feedButton(int a);
	// vlozi novou pozici mysi
	virtual void feedNewPos(float x, float y);
};


// stav >=4 je ve strehu:
const nSraz = 1, nSrazStreh = 2, nStrehSraz = 3, nStreh = 4, nPosun = 5, nOdsun = 6,
	  nStrehVypad = 7, nVypad = 8, nVypadStreh = 9;

const cNumAnims = 6;

class KeyboardInputDevice  {
protected:
	CMutex m_mutex;
	Model *m_model; // model k animaci
	float m_rootPos[3];
	Win32Timer m_timer;
	int m_stav;
	bool m_hnedZpatky; // mam se okamzite vratit z vypadu?

	KeyframeInterpolator *m_anim[cNumAnims];
	int anim_idx;
	float anim_time;

	CJudge *m_judge;
	int m_avatar;

protected:
	// pomocna funkce co prejde do jineho stavu pokud uz nadesel cas
	void testTime();

public:
	KeyboardInputDevice(Model *m, CJudge *judge, int avatar); // avatar je cislo avatara (pro rozhodciho)
	~KeyboardInputDevice();
	// krmi tridu informacemi a stisknuti/pusteni ("down") tlacitek
	void KeyboardEvent(bool down, int nChar, int repCnt = 0, int nFlags = 0);
	// nastavi polohu modelu podle daneho stavu a casu
	void setLegsAnimation();
	// zrusi probihajici animaci a vrati sermire na start:
	void resetAnimation();
	void setRootPos(const float rp[3]) 
	{ 
		CSingleLock lock(&m_mutex); lock.Lock();
		m_rootPos[0]=rp[0]; m_rootPos[1]=rp[1]; m_rootPos[2]=rp[2]; 
		lock.Unlock();
	}
	void getRootPos(float rp[3]) 
	{ 
		CSingleLock lock(&m_mutex); lock.Lock();
		rp[0]=m_rootPos[0]; rp[1]=m_rootPos[1]; rp[2]=m_rootPos[2]; 
		lock.Unlock();
	}

	void getAnimPars(int &aidx, float &atime) 
	{ 
		CSingleLock lock(&m_mutex); lock.Lock();
		aidx = anim_idx; atime = anim_time; 
		lock.Unlock();
	}
};

#endif