/* 
    KeyframeInterpolator.cpp

	Compute interpolated joint transformations. Including
	extended class for loading keyframe data from Milkshape.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include <assert.h>
#include "KeyframeInterpolator.h"
#include "Vector.h"
#include "Matrix.h"
#include <fstream.h>

KeyframeInterpolator::KeyframeInterpolator(Model *m) : m_model(m)  
{
	m_pJointKeyframes = new JointKeyframes[ m_model -> getNumJoints() ];
	m_totalTime = 0.0;
	m_lastTime = 0.0;
};

KeyframeInterpolator::~KeyframeInterpolator()  {
	for (int i=0;i<m_model -> getNumJoints();i++)
	{
		delete[] m_pJointKeyframes[i].m_pTranslationKeyframes;
        delete[] m_pJointKeyframes[i].m_pRotationKeyframes;
	}

	delete[] m_pJointKeyframes;
	m_pJointKeyframes = NULL;	
};

// write interpolated values for given time:
void KeyframeInterpolator::setTransform(double time)  
{
  if (time < m_lastTime)
  { // pozor, time lze jen zvetsovat, osetrim zde:
	  reset();	  
  }
  m_lastTime = time;  

  for ( int i = 0; i < m_model -> getNumJoints(); i++ )
	{
		float transVec[3];
		int frame;
		JointKeyframes *pJointKeyframes = &m_pJointKeyframes[i];
		Model::Joint *pJoint = &(m_model -> m_pJoints[i]);		

		if (( pJointKeyframes->m_numRotationKeyframes == 0 && 
			 pJointKeyframes->m_numTranslationKeyframes == 0 )) continue;

		/*if (i >= m_legs || i == 0)  */{ // hybe se se vsim (i s rukou)
			frame = pJointKeyframes->m_currentTranslationKeyframe;			
			while ( frame < pJointKeyframes->m_numTranslationKeyframes && pJointKeyframes->m_pTranslationKeyframes[frame].m_time < time )
			{
				frame++;
			}
			pJointKeyframes->m_currentTranslationKeyframe = frame;
			
			if ( frame == 0 )
				memcpy( transVec, pJointKeyframes->m_pTranslationKeyframes[0].m_parameter, sizeof ( float )*3 );				
			else if ( frame == pJointKeyframes->m_numTranslationKeyframes )
				memcpy( transVec, pJointKeyframes->m_pTranslationKeyframes[frame-1].m_parameter, sizeof ( float )*3 );				
			else
			{
				assert( frame > 0 && frame < pJointKeyframes->m_numTranslationKeyframes );
				
				const KeyframeInterpolator::Keyframe& curFrame = pJointKeyframes->m_pTranslationKeyframes[frame];
				const KeyframeInterpolator::Keyframe& prevFrame = pJointKeyframes->m_pTranslationKeyframes[frame-1];
				
				float timeDelta = curFrame.m_time-prevFrame.m_time;
				float interpValue = ( float )(( time-prevFrame.m_time )/timeDelta );

				assert( interpValue >= 0 && interpValue <= 1 );
				
				transVec[0] = prevFrame.m_parameter[0]+( curFrame.m_parameter[0]-prevFrame.m_parameter[0] )*interpValue;
				transVec[1] = prevFrame.m_parameter[1]+( curFrame.m_parameter[1]-prevFrame.m_parameter[1] )*interpValue;
				transVec[2] = prevFrame.m_parameter[2]+( curFrame.m_parameter[2]-prevFrame.m_parameter[2] )*interpValue; 
			}
			
			frame = pJointKeyframes->m_currentRotationKeyframe;
			while ( frame < pJointKeyframes->m_numRotationKeyframes && pJointKeyframes->m_pRotationKeyframes[frame].m_time < time )
			{
				frame++;
			}
			pJointKeyframes->m_currentRotationKeyframe = frame;
			
			if ( frame == 0 )
				pJoint->m_transform.setRotationRadians( pJointKeyframes->m_pRotationKeyframes[0].m_parameter );
				//memcpy ( pJoint->rot, pJointKeyframes->m_pRotationKeyframes[0].m_parameter, sizeof(float)*3);
				//Vector::set3(pJoint->rot, pJointKeyframes->m_pRotationKeyframes[0].m_parameter);
			else if ( frame == pJointKeyframes->m_numRotationKeyframes )
				pJoint->m_transform.setRotationRadians( pJointKeyframes->m_pRotationKeyframes[frame-1].m_parameter );
				//memcpy ( pJoint->rot, pJointKeyframes->m_pRotationKeyframes[frame-1].m_parameter, sizeof(float)*3);
				//Vector::set3(pJoint->rot, pJointKeyframes->m_pRotationKeyframes[frame-1].m_parameter);
			else
			{
				assert( frame > 0 && frame < pJointKeyframes->m_numRotationKeyframes );
				
				const KeyframeInterpolator::Keyframe& curFrame = pJointKeyframes->m_pRotationKeyframes[frame];
				const KeyframeInterpolator::Keyframe& prevFrame = pJointKeyframes->m_pRotationKeyframes[frame-1];
				
				float timeDelta = curFrame.m_time-prevFrame.m_time;
				float interpValue = ( float )(( time-prevFrame.m_time )/timeDelta );
								
				assert( interpValue >= 0 && interpValue <= 1 );
				
				Quaternion qPrev( prevFrame.m_parameter );
				Quaternion qCur( curFrame.m_parameter );
				Quaternion qFinal( qPrev, qCur, interpValue );
				pJoint->m_transform.setRotationQuaternion( qFinal );
/*
				float rotVec[3];
				
				rotVec[0] = prevFrame.m_parameter[0]+( curFrame.m_parameter[0]-prevFrame.m_parameter[0] )*interpValue;
				rotVec[1] = prevFrame.m_parameter[1]+( curFrame.m_parameter[1]-prevFrame.m_parameter[1] )*interpValue;
				rotVec[2] = prevFrame.m_parameter[2]+( curFrame.m_parameter[2]-prevFrame.m_parameter[2] )*interpValue;
				
				pJoint->m_transform.setRotationRadians( rotVec );
/**/
			}
				
			pJoint->m_transform.setTranslation( transVec );
        }		
	}
}

// pricte k "root_pos" rozdil mezi pozici rootu na konci a na zacatku pohybu
void KeyframeInterpolator::addRootTrans(float *root_pos)
{	
	float f[3];
	Vector::set3(f,m_pJointKeyframes[0].m_pTranslationKeyframes[
		m_pJointKeyframes[0].m_numTranslationKeyframes-1].m_parameter);
	Vector::sub3(f, m_pJointKeyframes[0].m_pTranslationKeyframes[0].m_parameter);
	Vector::add3(root_pos, f);
}

// nastavuje rotation/translation keyframe:
void KeyframeInterpolator::setJointKeyframe( int jointIndex, int keyframeIndex, float time, float *parameter, bool isRotation )
{
	assert( m_model -> getNumJoints() > jointIndex );

	Keyframe& keyframe = isRotation ? m_pJointKeyframes[jointIndex].m_pRotationKeyframes[keyframeIndex] :
		m_pJointKeyframes[jointIndex].m_pTranslationKeyframes[keyframeIndex];

	keyframe.m_jointIndex = jointIndex;
	keyframe.m_time = time;
	memcpy( keyframe.m_parameter, parameter, sizeof( float )*3 );
}

// sets current...Keyframes to zero
void KeyframeInterpolator::reset()
{
	for (int i=0; i<m_model -> getNumJoints(); i++)
	{
	  m_pJointKeyframes[i].m_currentRotationKeyframe = 0;
      m_pJointKeyframes[i].m_currentTranslationKeyframe = 0;
	}
}

// posune celou animaci tak, aby zacinala rootem v nule
void KeyframeInterpolator::normalizeRoot()
{	
	int fr = m_pJointKeyframes[0].m_numTranslationKeyframes;
	if (fr == 0) return;
	float start[3];
	Vector::set3(start, m_pJointKeyframes[0].m_pTranslationKeyframes[0].m_parameter);

	for (int n = 0; n < fr; n++)
	{
		Vector::sub3(m_pJointKeyframes[0].m_pTranslationKeyframes[n].m_parameter, start);
	}
}

#include "MilkshapeStructures.h"

// kopiruje keyframy v obracenem poradi (pozpatku pustena animace)
void KeyframeInterpolator::reverseKeyframeData( const KeyframeInterpolator &ki )
{
	m_totalTime = ki.m_totalTime; // stejne dlouha
//	m_legs = ki.m_legs;
	m_model = ki.m_model;

	for (int i = 0; i < m_model -> getNumJoints(); i++ )
	{
		int rot,trans;

		rot = m_pJointKeyframes[i].m_numRotationKeyframes = ki.m_pJointKeyframes[i].m_numRotationKeyframes;
		m_pJointKeyframes[i].m_pRotationKeyframes = new Keyframe[rot];
		trans = m_pJointKeyframes[i].m_numTranslationKeyframes = ki.m_pJointKeyframes[i].m_numTranslationKeyframes;
		m_pJointKeyframes[i].m_pTranslationKeyframes = new Keyframe[trans];

		int j;
		for ( j = 0; j < rot; j++ )
		{
			setJointKeyframe( i, rot-1-j, (float)m_totalTime - ki.m_pJointKeyframes[i].m_pRotationKeyframes[j].m_time, 
				ki.m_pJointKeyframes[i].m_pRotationKeyframes[j].m_parameter, true );
		}

		for ( j = 0; j < trans; j++ )
		{
			setJointKeyframe( i, trans-1-j, (float)m_totalTime - ki.m_pJointKeyframes[i].m_pTranslationKeyframes[j].m_time, 
				ki.m_pJointKeyframes[i].m_pTranslationKeyframes[j].m_parameter, false );

		}
	}
  normalizeRoot();
  reset();	
}

// loaduje keyframy ze souboru modelu:
bool MilkshapeKeyframeInterpolator::loadKeyframeData( const char *filename)
{
	ifstream inputFile( filename, ios::in | ios::binary | ios::nocreate );
	if ( inputFile.fail())
		return false;	// "Couldn't open the model file."

	char pathTemp[PATH_MAX+1];
	int pathLength;
	for ( pathLength = strlen( filename ); pathLength--; ) {
		if ( filename[pathLength] == '/' || filename[pathLength] == '\\' ) {
			break;
		}
	}
	strncpy( pathTemp, filename, pathLength );

	inputFile.seekg( 0, ios::end );
	long fileSize = inputFile.tellg();
	inputFile.seekg( 0, ios::beg );

	byte *pBuffer = new byte[fileSize];
	inputFile.read( pBuffer, fileSize );
	inputFile.close();

	const byte *pPtr = pBuffer;
	MS3DHeader *pHeader = ( MS3DHeader* )pPtr;
	pPtr += sizeof( MS3DHeader );

	if ( strncmp( pHeader->m_ID, "MS3D000000", 10 ) != 0 )
		return false; // "Not a valid Milkshape3D model file."

	if ( pHeader->m_version < 3 || pHeader->m_version > 4 )
		return false; // "Unhandled file version. Only Milkshape3D Version 1.3 and 1.4 is supported." );

	int nVertices = *( word* )pPtr; 
//	m_numVertices = nVertices;
//	m_pVertices = new Vertex[nVertices];
	pPtr += sizeof( word );

	int i;
	pPtr += sizeof( MS3DVertex ) * nVertices; // preskoc info o vrcholech

	int nTriangles = *( word* )pPtr;
//	m_numTriangles = nTriangles;
//	m_pTriangles = new Triangle[nTriangles];
	pPtr += sizeof( word );

	pPtr += sizeof( MS3DTriangle ) * nTriangles;

	int nGroups = *( word* )pPtr;
//	m_numMeshes = nGroups;
//	m_pMeshes = new Mesh[nGroups];
	pPtr += sizeof( word );

	for ( i = 0; i < nGroups; i++ )
	{
		pPtr += sizeof( byte );	// flags
		pPtr += 32;				// name

		word nTriangles = *( word* )pPtr;
		pPtr += sizeof( word );
		pPtr += sizeof( word ) * nTriangles;

		pPtr += sizeof( char );
	}

	int nMaterials = *( word* )pPtr;
//	m_numMaterials = nMaterials;
//	m_pMaterials = new Material[nMaterials];
	pPtr += sizeof( word );
	pPtr += sizeof( MS3DMaterial) * nMaterials;

	// Load Skeleton Definiton

	float animFPS = *( float* )pPtr;
	pPtr += sizeof( float );

	// skip currentTime
	pPtr += sizeof( float );

	int totalFrames = *( int* )pPtr;
	pPtr += sizeof( int );

	m_totalTime = totalFrames*1000.0/animFPS;

	int m_numJoints = *( word* )pPtr;
	assert( m_numJoints == m_model -> getNumJoints() );
	pPtr += sizeof( word );

	/* obsoletni: zjistovani indexu RightLeg1... nove hybu se vsim
	m_legs = -1;
    if (legs_only)  {
	  //m_pJoints = new Joint[m_numJoints];

	  struct JointNameListRec
	  {
	  	int m_jointIndex;
		const char *m_pName;
	  };

	  const byte *pTempPtr = pPtr;

	  //JointNameListRec *pNameList = new JointNameListRec[m_numJoints];
	  for ( i = 0; i < m_numJoints; i++ )
	  {
		MS3DJoint *pJoint = ( MS3DJoint* )pTempPtr;
		pTempPtr += sizeof( MS3DJoint );
		pTempPtr += sizeof( MS3DKeyframe )*( pJoint->m_numRotationKeyframes+pJoint->m_numTranslationKeyframes );

		if (strcmp(pJoint->m_name, "RightLeg1")==0)  {
			m_legs = i;
		}
		
		//pNameList[i].m_jointIndex = i;
		//pNameList[i].m_pName = pJoint->m_name;
	  }
	  assert(m_legs >= 0); // nenalezna 'RightLeg1' !
	} 
	*/

	for ( i = 0; i < m_numJoints; i++ )
	{
		MS3DJoint *pJoint = ( MS3DJoint* )pPtr;
		pPtr += sizeof( MS3DJoint );

		/*
		int j, parentIndex = -1;
		if ( strlen( pJoint->m_parentName ) > 0 )
		{
			for ( j = 0; j < m_numJoints; j++ )
			{
				if ( stricmp( pNameList[j].m_pName, pJoint->m_parentName ) == 0 )
				{
					parentIndex = pNameList[j].m_jointIndex;
					break;
				}
			}
			if ( parentIndex == -1 ) // Unable to find parent bone in MS3D file
				return false;
		}

		memcpy( m_pJoints[i].m_localRotation, pJoint->m_rotation, sizeof( float )*3 );
		memcpy( m_pJoints[i].m_localTranslation, pJoint->m_translation, sizeof( float )*3 );
		m_pJoints[i].m_parent = parentIndex;
		*/
		
		m_pJointKeyframes[i].m_numRotationKeyframes = pJoint->m_numRotationKeyframes;
		m_pJointKeyframes[i].m_pRotationKeyframes = new Keyframe[pJoint->m_numRotationKeyframes];
		m_pJointKeyframes[i].m_numTranslationKeyframes = pJoint->m_numTranslationKeyframes;
		m_pJointKeyframes[i].m_pTranslationKeyframes = new Keyframe[pJoint->m_numTranslationKeyframes];

		int j;
		for ( j = 0; j < m_pJointKeyframes[i].m_numRotationKeyframes; j++ )
		{
			MS3DKeyframe *pKeyframe = ( MS3DKeyframe* )pPtr;
			pPtr += sizeof( MS3DKeyframe );

			setJointKeyframe( i, j, pKeyframe->m_time*1000.0f, pKeyframe->m_parameter, true );
		}

		for ( j = 0; j < m_pJointKeyframes[i].m_numTranslationKeyframes; j++ )
		{
			MS3DKeyframe *pKeyframe = ( MS3DKeyframe* )pPtr;
			pPtr += sizeof( MS3DKeyframe );

			setJointKeyframe( i, j, pKeyframe->m_time*1000.0f, pKeyframe->m_parameter, false );
		}
	}
//	delete[] pNameList;

//	setupJoints();
// { end NEW }
	delete[] pBuffer; 

	normalizeRoot();
	reset();

	return true;
}

