/* 
    LocalRemoteModel.cpp

	Class RemoteModel drives posture of remote (opponent) model using
	ModelState transferred by network.
	Class LocalModel drives posture of local (my) model using input
	devices.


    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "LocalRemoteModel.h"
#include "MilkshapeModel.h"
#include "ColDet.h"
#include <assert.h>
#include "ColResp.h"
#include <direct.h>
#include "paths.h"

ModelState zeroModelState;

void FencerModel::initModel(int typ, int quality, int kDOP)
{
	m_typ = typ; m_kDOP = kDOP;
	m_model = new MilkshapeModel();
	char *mfile[2] = {"model.ms3d", "modelBl.ms3d"};
	if ( m_model->loadModelData( DATA_PATH + mfile[typ] ) == false )		// Loads The Model And Checks For Errors
	{
		::MessageBox( NULL, "Couldn't load the model file 'model.ms3d'", "Error", MB_OK | MB_ICONERROR );		
		char tst[128];	_getcwd(tst, 128);::MessageBox(0, tst, "Working directory:", MB_OK);
		exit(-1);
	}
	if (m_model->m_swordholder == -1 || m_model->m_ikj[HAND] == -1 ||
		m_model->m_ikj[ELBOW] == -1 || m_model->m_ikj[SHOULDER] == -1 ||
		m_model->m_midfinger == -1 || m_model->m_headJoint == -1) 
		TRACE0("Not found obligatory joints in model.ms3d!\n");	

	if ( m_model -> loadWeapon( DATA_PATH + "mec.ms3d") == false )
	{
		::MessageBox( NULL, "Couldn't load the weapon model file 'mec.ms3d'", "Error", MB_OK | MB_ICONERROR );		
		char tst[128];	_getcwd(tst, 128);::MessageBox(0, tst, "Working directory:", MB_OK);
		exit(-1);
	}	
	if ( m_model -> loadNormals(DATA_PATH + "normaly.txt") == false )
	{
		::MessageBox( NULL, "Couldn't load the normals file 'normaly.txt'", "Error", MB_OK | MB_ICONERROR );
		char tst[128];	_getcwd(tst, 128);::MessageBox(0, tst, "Working directory:", MB_OK);
		exit(-1);
	}
	// nutnte jen pro vertex blending (kvuli srovnavani):
	if ( m_model->loadWeights(DATA_PATH + "vahy.txt") == false )
	{
		::MessageBox( NULL, "Couldn't load the vertex blending weights file 'vahy.txt'", "Error", MB_OK | MB_ICONERROR );
		char tst[128];	_getcwd(tst, 128);::MessageBox(0, tst, "Working directory:", MB_OK);
		exit(-1);
	}

	m_quality = quality;
	if (quality)
	{
		m_model->subdivMesh1k4(m_model->m_ikj[HAND], m_model->m_ikj[ELBOW]);
		m_model->subdivMesh1k4(m_model->m_ikj[SHOULDER]);
		m_model->subdivMesh1k4(m_model->m_ikj[ELBOW]);
		m_model->subdivMesh1k4(m_model->m_ikj[HAND]);
		m_model->subdivMesh1k4(4);
	}

	// nastaveni nezasahovych trojuhelniku (nemusi byt nutne pred inicializaci CD):
	m_model->setNonhitTriangles();

	//m_model->saveModelData(DATA_PATH + "saved.txt");

	// pro ucely srovnavani metod animace:
	//m_model->m_renderingMethod = BASIC_SA;

	m_cdm = new CDModel(m_model, kDOP, false);
	m_cdw = new CDModel(m_model->getWeapon(), kDOP, true);
}

FencerModel::FencerModel(int typ, const float rp[3], int quality, int kDOP)
{
	if (rp) Vector::set3(m_state.root_pos, rp);
	else 
	{
		m_state.root_pos[0] = 0;m_state.root_pos[1] = 0;m_state.root_pos[2] = 0;
	}

	initModel(typ, quality, kDOP);
}

//DEL void FencerModel::changeQuality(int quality)
//DEL {
//DEL 	delete m_cdm;
//DEL 	delete m_cdw;
//DEL 	delete m_model;
//DEL 	m_quality = quality;
//DEL 	initModel(m_typ, m_quality, m_kDOP);
//DEL }

void FencerModel::setRootTrans(int idx)
{
	Matrix rt;
	if (idx==1) // sermir vlevo, diva se vpravo
	{
		float rot[3] = {0, 90, 0};
		float pos[3] = {0, 0, 70};
		rt.setRotationDegrees(rot);
		rt.setTranslation(pos);
	}
	else if (idx==2) // obracene
	{
		float rot[3] = {0, -90, 0};
		float pos[3] = {0, 0, -70};
		rt.setRotationDegrees(rot);
		rt.setTranslation(pos);
	}
	m_model->setRootTransform(rt);
	if (m_model->getWeapon()) m_model->getWeapon()->setRootTransform(rt);
}

// prepocita finalni matice (s aktualnim root_pos);
void FencerModel::setFinalMatrices()
{
	m_model->setFinalMatrices(m_state.root_pos);
}

// updateuje stav modelu novou pozici a orientaci zbrane a lokte (kvuli zmenam z colResp):
void FencerModel::updateModelState(const Matrix &nEE_or, const float nEE_pos[3], const float nelbow_pos[3])
{
	m_state.EE_or.set(nEE_or.getMatrix());
	cpvector(m_state.EE_pos, nEE_pos);
	cpvector(m_state.elbow_pos, nelbow_pos);
}

FencerModel::~FencerModel()
{
	delete m_cdm;
	delete m_cdw;
	delete m_model;
}

RemoteModel::RemoteModel(int typ, const float rp[3], int quality, int kDOP) : FencerModel(typ, rp, quality, kDOP)
{
	m_IKanim = new IKAN(m_model);

	m_anim[0] = new MilkshapeKeyframeInterpolator(m_model);
	if ( m_anim[0]->loadKeyframeData( DATA_PATH + "model.ms3d" ) == false )
	{ 
		::MessageBox( NULL, "Couldn't load the keyframe file 'model.ms3d", "Error", MB_OK | MB_ICONERROR ); 
	}
	m_anim[1] = new MilkshapeKeyframeInterpolator(m_model);
	m_anim[1] -> reverseKeyframeData( *m_anim[0] );
	m_anim[2] = new MilkshapeKeyframeInterpolator(m_model);
	if ( m_anim[2]->loadKeyframeData( DATA_PATH + "posun.ms3d" ) == false )
	{
		::MessageBox( NULL, "Couldn't load the keyframe file 'posun.ms3d", "Error", MB_OK | MB_ICONERROR );		
	}
	m_anim[3] = new MilkshapeKeyframeInterpolator(m_model);
	m_anim[3] -> reverseKeyframeData( *m_anim[2] );
	m_anim[4] = new MilkshapeKeyframeInterpolator(m_model);
	if ( m_anim[4]->loadKeyframeData( DATA_PATH + "vypad.ms3d" ) == false )
	{
		::MessageBox( NULL, "Couldn't load the keyframe file 'posun.ms3d", "Error", MB_OK | MB_ICONERROR );		
	}
	m_anim[5] = new MilkshapeKeyframeInterpolator(m_model);
	m_anim[5] -> reverseKeyframeData( *m_anim[4] );
}

RemoteModel::~RemoteModel()
{
	delete m_IKanim;
	for (int n=0; n<cNumAnims; n++) delete m_anim[n];
}

// nastavuje se podle daneho stavu
void RemoteModel::changedModelState()
{
	assert(m_state.anim_idx >= 0 && m_state.anim_idx < cNumAnims);

	m_anim[m_state.anim_idx] -> setTransform(m_state.anim_time);	
	m_model ->setFinalMatrices(m_state.root_pos); // pred volanim IK by mely bejt nastaveny	
	Vector EE(m_state.EE_pos);
	Matrix EE_or(m_state.EE_or); // pozor, meni ji to
	float EEpos[3];
	EE.get(EEpos);
/*
	// applyDisp pracuje relativne k rootu, m_state.EE_pos a or mam taky vzhledem k rootu,
	// ale IK to musim predat ve svetovych souradnicich (prenasobene zleva root.final):
	if (cr) cr->applyDisplacements(1, EE_or, EEpos, m_state.elbow_pos);
	// vratit EE_or a EEpos do stavu modelu (kvuli sejvovani):
	cpvector(m_state.EE_pos, EEpos);
	m_state.EE_or.set(EE_or.getMatrix());
*/
	// pridam transformaci root.final (setFinalMatrices spocitano driv):
	Matrix EEmat;
	EEmat.set(EE_or.getMatrix());
	EEmat.setTranslation(EEpos);
	EEmat.preMultiply(m_model->m_pJoints[0].m_final);
	EE_or.set(EEmat.getMatrix()); EE_or.setTranslation(zero3);
	EEmat.getTranslation(EEpos);
	EE.set(EEpos);
	Vector::set3(m_IKanim -> elbowPos, m_state.elbow_pos);	
	m_IKanim->setGoal(EE, EEmat, m_IKanim->elbowPos); // pozor, meni je to!
	m_model -> setFinalMatrices(m_state.root_pos); //propagujeme transformace
}

// nastavuje se podle vstupnich zarizeni
// h je cislo hrace - POZOR, hrac cislo h musi byt stejny v CChildView i v ColResp!
void LocalModel::updateModel(int h, void *vcr) // a nastavit m_state !
{
	ColResp *cr = (ColResp *)vcr; // kvuli zavislostem headeru... fuj
	bool changed;

	//float legStart[3], legTrans[3];
	//m_model->m_pJoints[0].m_final.getTranslation(legStart);

	m_kid -> setLegsAnimation();
	m_kid -> getAnimPars(m_state.anim_idx, m_state.anim_time);	
	m_kid -> getRootPos(m_state.root_pos);
	setFinalMatrices();

	//m_model->m_pJoints[0].m_final.getTranslation(legTrans);
	//Vector::sub3(legTrans, legStart);
	//TRACE("legTrans: %f %f %f\n", legTrans[0], legTrans[1], legTrans[2]);

	Vector EE;
	m_mid->getGoal(m_state.EE_or, EE);
	/*{ // spocitat ciste MIDovou polohu (kvuli getSHFfromMID);

		m_midSHF.set(m_state.EE_or.getMatrix());
		m_midSHF.setTranslation(EE.getVector());	
		m_midSHF.preMultiply(m_model->m_pJoints[0].m_final);

		// jeste to musim prohnat IK, protoze ne vsechny polohy z MIDu jsou dosazitelne (1.3.03):
		Matrix EEor; Vector EEpos;
		EEor = m_midSHF; EEor.setTranslation(zero3);
		float pom[3];
		m_midSHF.getTranslation(pom);
		EEpos.set(pom);
		m_IKanim->setGoal(EEpos, EEor, m_IKanim->elbowPos);
		setFinalMatrices();
		m_midSHF.set(m_model->getSHF());
	}/**/
	EE.get(m_state.EE_pos); // potrebuju to z MIDu nakopirovat do m_state
	m_toughness = m_mid->getToughness();
	m_mid->getElbow(m_state.elbow_pos);
	//m_mid->getElbow(m_IKanim -> elbowPos);
	// predpokladam ze z MIDu leze EE vzhledem k root.finalu (viz konvence)
	// aplikuju displacementy (pracuje vzhledem k rootu):
	//TRACE("PRED applyDisp je m_state.EE_pos = %f %f %f\n", m_state.EE_pos[0], m_state.EE_pos[1], m_state.EE_pos[2]);
	if (cr) changed = cr->applyDisplacements(h, m_state.EE_or, m_state.EE_pos, m_state.elbow_pos);
	//TRACE("PO applyDisp je m_state.EE_pos = %f %f %f\n", m_state.EE_pos[0], m_state.EE_pos[1], m_state.EE_pos[2]);
	// posunout EE do aktualni polohy modelu (tj. pridat transformaci rootu):
	// (to uz se neuklada do model state, tam je to vzhledem k root.finalu)
	Matrix EEmat;
	float actEE_pos[3];	
	EEmat.set(m_state.EE_or.getMatrix());
	EEmat.setTranslation(m_state.EE_pos);	
	EEmat.preMultiply(m_model->m_pJoints[0].m_final);
	EEmat.getTranslation(actEE_pos);
	//TRACE("LocalModel::updateModel hlasi actEE_pos = [%f, %f, %f]\n", actEE_pos[0], actEE_pos[1], actEE_pos[2]);
	EEmat.setTranslation(zero3);	
	EE.set(actEE_pos);
	// m_state.elbow_pos se meni v applyDisplacements
	cpvector(m_IKanim->elbowPos, m_state.elbow_pos); 
	m_IKanim->setGoal(EE, EEmat, m_IKanim->elbowPos);
	setFinalMatrices();
	// kvuli getSHFfromMID tady ulozim vysledne SHF vznikle jen z MIDu a applyDisplac
	m_midSHF.set(m_model->getSHF());

/* testik jestli se pozice SHF timhletim nezmenila: (zmenit by se mela jenom applyDisplacem)
	Matrix old, nov;
	nov.set(m_model->getSHF());
	float tr[3], pom[3];
	getmidSHF(old);
	old.getTranslation(tr);
	nov.getTranslation(pom);
	Vector::sub3(tr, pom);
	float vel = norm(tr);
	TRACE("LocalModel::updateModel - vel = %f\n", vel);	
/**/

/* test velikosti chyby: -- pozor, chyba muze vzniknout taky kvuli tomu, ze z MIDu lezou
   i nedosahnutelne pozice a IKAN si to musi skalovat...
	Matrix SHFnew(m_model->m_pJoints[m_model->m_swordholder].m_final);
	float hm[3],hm2[3];
	cpvector(hm, EE.getVector());
	SHFnew.getTranslation(hm2);
	Vector::sub3(hm, hm2);
	TRACE("error %f\n", norm(hm));
/**/
//testik:
/*
	Matrix tst(m_model->m_pJoints[m_model->m_swordholder].m_final);
		
	Matrix EEor(tst);		
	float nul[3] = {0,0,0}; EEor.setTranslation(nul);
	tst.getTranslation(nul);
	Vector EEpos(nul);
	
	m_IKanim->setGoal(EEpos, EEor);

	Matrix SHF(m_model->m_pJoints[m_model->m_swordholder].m_final);

	//m_model -> setFinalMatrices(m_state.root_pos); //propagujeme transformace
	setFinalMatrices();

	Matrix SHFnew(m_model->m_pJoints[m_model->m_swordholder].m_final);
/**/
}

LocalModel::LocalModel(CJudge *judge, int typ, const float rp[3], int quality, bool loadMS3D, int kDOP) : FencerModel(typ, rp, quality, kDOP)
{
	m_IKanim = new IKAN(m_model);

	m_kid = new KeyboardInputDevice(m_model, judge, typ);

	m_mid = new ActionMID(m_model, 3,3, 3,3, 5,3); // naprahy, seky, kryty
//	m_mid->setPars(CRect(0, 0, 400, 400), 40.0);

	if (loadMS3D)
	{ // nacti z MS3D souboru a uloz do souboru naprahy.dat, seky.dat, kryty.dat
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\levy kanec.ms3d", 0, 0);
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\levy trpaslik.ms3d", 0, 1);
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\levy byk.ms3d", 0, 2);
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\vyzva.ms3d", 1, 0);
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\primy bod.ms3d", 1, 1);
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\strecha.ms3d", 1, 2);
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\pravy kanec.ms3d", 2, 0);
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\pravy trpaslik.ms3d", 2, 1);
		m_mid -> m_naprahy.loadKeyframe(DATA_PATH + "keyframes\\pravy byk.ms3d", 2, 2);	
		
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt prima.ms3d", 0, 0);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt bod prima.ms3d", 1, 0);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt dolni.ms3d", 2, 0);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt bod sekonda.ms3d", 3, 0);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt sekonda.ms3d", 4, 0);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt kvarta.ms3d", 0, 1);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt bod kvarta.ms3d", 1, 1);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt bodu.ms3d", 2, 1);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt bod terca.ms3d", 3, 1);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt terca.ms3d", 4, 1);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt LB.ms3d", 0, 2);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt sixta.ms3d", 1, 2);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt stred 56.ms3d", 2, 2);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt kvinta.ms3d", 3, 2);
		m_mid -> m_kryty.loadKeyframe(DATA_PATH + "keyframes\\kryt PB.ms3d", 4, 2);	
		
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek LK.ms3d", 0, 0);
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek LT.ms3d", 0, 1);
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek LB.ms3d", 0, 2);
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek dolni.ms3d", 1, 0);
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek bod.ms3d", 1, 1);
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek strecha.ms3d", 1, 2);
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek PK.ms3d", 2, 0);
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek PT.ms3d", 2, 1);
		m_mid -> m_seky.loadKeyframe(DATA_PATH + "keyframes\\sek PB.ms3d", 2, 2);
				
		m_mid->m_naprahy.saveAllKeyframes(DATA_PATH + "naprahy.dat");
		m_mid->m_seky.saveAllKeyframes(DATA_PATH + "seky.dat");
		m_mid->m_kryty.saveAllKeyframes(DATA_PATH + "kryty.dat");
	}
	else
	{
		if (!m_mid->m_naprahy.loadAllKeyframes(DATA_PATH + "naprahy.dat"))
			MessageBox(NULL, "Couldn't load the keyframe data 'naprahy.dat'", "Error", MB_OK | MB_ICONERROR );
		if (!m_mid->m_seky.loadAllKeyframes(DATA_PATH + "seky.dat"))
			MessageBox(NULL, "Couldn't load the keyframe data 'seky.dat'", "Error", MB_OK | MB_ICONERROR );
		if (!m_mid->m_kryty.loadAllKeyframes(DATA_PATH + "kryty.dat"))
			MessageBox(NULL, "Couldn't load the keyframe data 'kryty.dat'", "Error", MB_OK | MB_ICONERROR );
	}

	if (!m_mid -> m_naprahy.loadElbowPos(DATA_PATH + "loket naprahy.txt"))	
	  ::MessageBox( NULL, "Couldn't load the keyframe data 'loket naprahy.txt'", "Error", MB_OK | MB_ICONERROR );
	if (!m_mid -> m_kryty.loadElbowPos(DATA_PATH + "loket kryty.txt"))	
	  ::MessageBox( NULL, "Couldn't load the keyframe data 'loket kryty.txt'", "Error", MB_OK | MB_ICONERROR );	
	if (!m_mid -> m_seky.loadElbowPos(DATA_PATH + "loket seky.txt"))
	  ::MessageBox( NULL, "Couldn't load the keyframe data 'loket kryty.txt'", "Error", MB_OK | MB_ICONERROR );	
}

LocalModel::~LocalModel()
{
	delete m_IKanim;
	delete m_kid;
	delete m_mid;
}

/*
class RemoteModel
{
protected:

	Model *m_pModel; 
	KeyframeInterpolator *m_anim[cNumAnims];
	IK *IKanim;

public:
	RemoteModel(Model *m);
	void supplyNewState(const ModelState state);
};
*/

/*
// vrati aktualni pozici a orientaci hlavy ve svetovych souradnicich:
void FencerModel::getHead(Matrix &head)
{
	setFinalMatrices();
	head.set( m_model->m_pJoints[m_model->m_headJoint].m_final.getMatrix() );
}
*/

// vrati aktualni pozici kloubu joint ve svetovych souradnicich:
void FencerModel::getJointPos(float pos[3], int joint)
{
	setFinalMatrices();
	//root.set( m_model->m_pJoints[0].m_final.getMatrix() );
	m_model->m_pJoints[joint].m_final.getTranslation(pos);
}
