/* 
    LocalRemoteModel.h

	Class RemoteModel drives posture of remote (opponent) model using
	ModelState transferred by network.
	Class LocalModel drives posture of local (my) model using input
	devices.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#ifndef LOCALREMOTEMODEL_H
#define LOCALREMOTEMODEL_H

#include "Model.h"
#include "KeyframeInterpolator.h"
#include "InputDevice.h"
#include "ikan.h"
#include "Matrix.h"
#include "Judge.h"

const int cDefault_kDOP = 6;

// Canonical 
typedef class CModelState
{
public:
	float root_pos[3]; // posunuti rootu
	int anim_idx; // index animace ktera probiha
	float anim_time; // jeji posunuti
	Matrix EE_or; // orientace ruky
	float EE_pos[3]; // pozice ruky; orientace a pozice ruky se udava relativne k root.finalu
	float elbow_pos[3]; // pozice lokte

	bool operator==(const class CModelState &ms2)
	{
		bool eq = true;
		for (int n=0; n<sizeof(CModelState); n++)
		{
			if (((char *)this)[n] != ((const char *)&ms2)[n]) 
			{
				eq=false;
				break;
			}
		}
		return eq;
	}
} ModelState;

extern ModelState zeroModelState; // nulovy (pro inicializaci)

// standard shell of model
class FencerModel
{
protected:
	ModelState m_state;	
	void initModel(int typ, int quality = 0, int kDOP = cDefault_kDOP);
public:	
	Model *m_model;
	CDModel *m_cdm, *m_cdw; // obalka pro model a jeho zbran
	float m_toughness; // aktualni pevnost drzeni zbrane
	int	m_quality, m_typ, m_kDOP;

	FencerModel(int typ, const float rp[3] = NULL, int quality = 1, int kDOP = cDefault_kDOP);
	virtual ~FencerModel();
	// updateuje stav modelu novou pozici a orientaci zbrane a lokte (kvuli zmenam z colResp):
	void updateModelState(const Matrix &nEE_or, const float nEE_pos[3], const float nelbow_pos[3]);
	// vrati stav modelu jako pole charu
	const char *getModelState() { return (const char *)&m_state; }
	// nastavi stav modelu z pole charu
	void setModelState(const char *st) { memcpy(&m_state, st, sizeof(ModelState)); }
	// resetuje stav modelu:
//DEL 	void resetModelState() { memcpy(&m_state, &zeroModelState, sizeof(ModelState)); }
	// nastavi root transofrmaci: 0 = identita, 1 = vlevo, 2 = vpravo
	void setRootTrans(int idx);
	// prepocita finalni matice (s aktualnim root_pos);
	void setFinalMatrices();
	// vrati aktualni pozici kloubu joint ve svetovych souradnicich:
	void getJointPos(float pos[3], int joint);	
};


class RemoteModel : public FencerModel
{
//protected: (jen kvuli blbosti) -- 6.9.02
public:

	KeyframeInterpolator *m_anim[cNumAnims];
	IKAN *m_IKanim;

public:
	RemoteModel(int typ, const float rp[3] = NULL, int quality = 1, int kDOP = cDefault_kDOP);
	virtual ~RemoteModel();
	// prevede zmeny z m_state (zmeneneho napr. fci setModelState) do modelu
	void changedModelState();
};


class LocalModel : public FencerModel
{
public:	
	IKAN *m_IKanim;

public:
	KeyboardInputDevice		*m_kid;
	ActionMID *m_mid;	

	// typ je cislo avatara
	LocalModel(CJudge *judge, int typ, const float rp[3] = NULL, int quality = 1, bool loadMS3D = false, int kDOP = cDefault_kDOP);
	virtual ~LocalModel();
	// precte nove hodnoty ze vstupnich zarizeni a aktualizuje stav modelu
	// h je cislo hrace - POZOR, hrac cislo h musi byt stejny v CChildView i v ColResp!
	void updateModel(int h, void *vcr); // (taky nastavuje m_state)

	// vraci SHF takove, jake by bylo jenom podle MIDu a applyDisplac (presneji podle jeho 
	// stavu v okamziku posledniho updateModelu), tedy bez ohledu na ColResp
	// pozor, na rozdil od MIDu to vraci ve svetovych souradnicich:
	void getmidSHF(Matrix &SHF) { SHF = m_midSHF; }

protected:
	Matrix m_midSHF;  // jaka by byla SHF, kdyby to bylo jen na MIDu (nastavuje se v updateModel)
};

#endif