/*
	Matrix.h

		Matrix manipulation.

	Author:	Brett Porter
	Email: brettporter@yahoo.com
	Website: http://rsn.gamedev.net/pl3d
	Copyright (C)2000, 2001, Brett Porter. All Rights Reserved.
	This source code is released under the LGPL. See license.txt for details.

	Created: 9 December 2000
	Last Edited: 8 January 2001

	Please see the file ChangeLog.html for a revision history.
*/

/*
	Modifications by Ladislav Kavan, 2002
*/

#ifndef PORTALIB3D_MATRIX_HXX
#define PORTALIB3D_MATRIX_HXX

#include "Quaternion.h"
#include <string.h>

class Vector;

const float zero3[3] = {0,0,0};

class Matrix
{
	public:
		/*	Constructor. */
		Matrix();

		/*	Destructor. */
		~Matrix();

		/*	Set to identity. */
		void loadIdentity();

		void loadZero();

		/*	Set the values of the matrix. */
		void set( const float *matrix );

		/*	Post-multiply by another matrix. */
		void postMultiply( const Matrix& matrix );

		/*	Pre-multiply by another matrix. */
		void preMultiply( const Matrix& matrix );

		/*	Pre-multiply by another matrix. Translation part remains intact. */
		void preRotMultiply( const Matrix& matrix );

		/*	Set the translation of the current matrix. Will erase any previous values. */
		void setTranslation( const float *translation );

		/* Get translation from the matrix. */
		void getTranslation( float *translation ) const;

		/* Add the translation to the current matrix. */
		void addTranslation( const float *translation );

		/*	Set the inverse translation of the current matrix. Will erase any previous values. */
		void setInverseTranslation( const float *translation );

		/*	Make a rotation matrix from Euler angles. The 4th row and column are unmodified. */
		void setRotationRadians( const float *angles );

		/*	Make a rotation matrix from Euler angles. The 4th row and column are unmodified. */
		void setRotationDegrees( const float *angles );

		/*	Make a rotation matrix from a quaternion. The 4th row and column are unmodified. */
		void setRotationQuaternion( const Quaternion& quat );

		/*	Make a rotation matrix from axis and angle. The 4th row and column are unmodified. */
		void setRotationAxisAngle(const float axis[3], float angle);

		/*	Make an inverted rotation matrix from Euler angles. The 4th row and column are unmodified. */
		void setInverseRotationRadians( const float *angles );

		/*	Make an inverted rotation matrix from Euler angles. The 4th row and column are unmodified. */
		void setInverseRotationDegrees( const float *angles );

		/*	Get the matrix data. */
		const float *getMatrix() const { return m_matrix; }

		/*  Get orientation and translation component */
		void getOrPos( Matrix &or, Vector &pos) const;

		/*  Set orientation and translation component */
		void setOrPos( const Matrix &or, const Vector &pos);

		/* Get column to 3 float array */
		void getColumn(int idx, float *v);

		/* Set column to 3 float array */
		void setColumn(int idx, const float v[3]);

		/*	Translate a vector by the inverse of the translation part of this matrix. */
		void inverseTranslateVect( float *pVect ) const;

		/*	Translate a vector by the translation part of this matrix. */
		void TranslateVect( float *pVect ) const;

		/*	Rotate a vector by the inverse of the rotation part of this matrix. */
		void inverseRotateVect( float *pVect ) const;

		/*	Rotate a vector by the rotation part of this matrix. */
		void RotateVect( float *pVect ) const;

		/* Returns inverse rotation matrix (i.e. transposed) (only 3x3) */
		void inverseRotateMatrix( Matrix &ir ) const;

		/* Return inverse homogenous matrix */
		void inverseHomogenousMatrix( Matrix &ir ) const;

		/* Multiplication by scalar. */
		void mulScal( float s, Matrix &res ) const;

		/* Add matrix */
		void add( const Matrix m );

		/* SLERP: res = (1-t) * this  + t * m2 */
		// obracene vyzada interpolaci pres tupy uhel:
		void SLERP( Matrix &res, const Matrix &m2, float t, bool obracene = false);

		/* inverze, nepredpoklada nic o ortogonalite */
		bool invert();

		/* extract axis and angle */
		void axisangle( float axis[3], float &angle);

		/* Sum of squared differences */
		float l2norm( const Matrix m);

	private:
		//	Matrix data, stored in column-major order
		float m_matrix[16];
};

inline void Matrix::set( const float *matrix )
{
	memcpy( m_matrix, matrix, sizeof( float )*16 );
}

inline void Matrix::loadIdentity()
{
	memset( m_matrix, 0, sizeof( float )*16 );
	m_matrix[0] = m_matrix[5] = m_matrix[10] = m_matrix[15] = 1;
}

inline void Matrix::loadZero()
{
	memset( m_matrix, 0, sizeof( float )*16 );
}

inline void Matrix::inverseRotateVect( float *pVect ) const
{ // matice rotace je ortogonalni
	float vec[3];

	vec[0] = pVect[0]*m_matrix[0]+pVect[1]*m_matrix[1]+pVect[2]*m_matrix[2];
	vec[1] = pVect[0]*m_matrix[4]+pVect[1]*m_matrix[5]+pVect[2]*m_matrix[6];
	vec[2] = pVect[0]*m_matrix[8]+pVect[1]*m_matrix[9]+pVect[2]*m_matrix[10];

	memcpy( pVect, vec, sizeof( float )*3 );
}

inline void Matrix::RotateVect( float *pVect ) const
{
	float vec[3];

	vec[0] = pVect[0]*m_matrix[0]+pVect[1]*m_matrix[4]+pVect[2]*m_matrix[8];
	vec[1] = pVect[0]*m_matrix[1]+pVect[1]*m_matrix[5]+pVect[2]*m_matrix[9];
	vec[2] = pVect[0]*m_matrix[2]+pVect[1]*m_matrix[6]+pVect[2]*m_matrix[10];

	memcpy( pVect, vec, sizeof( float )*3 );
}

inline void Matrix::inverseTranslateVect( float *pVect ) const
{
	pVect[0] = pVect[0]-m_matrix[12];
	pVect[1] = pVect[1]-m_matrix[13];
	pVect[2] = pVect[2]-m_matrix[14];
}

inline void Matrix::TranslateVect( float *pVect ) const
{
	pVect[0] = pVect[0]+m_matrix[12];
	pVect[1] = pVect[1]+m_matrix[13];
	pVect[2] = pVect[2]+m_matrix[14];
}

void matrix_test();

#endif // ndef PORTALIB3D_MATRIX_HXX
