/*
	MilkshapeModel.cpp

		Loads and renders a Milkshape3D model. 

	Author:	Brett Porter
	Email: brettporter@yahoo.com
	Website: http://www.geocities.com/brettporter/
	Copyright (C)2000, Brett Porter. All Rights Reserved.

	This file may be used only as long as this copyright notice remains intact.
*/

/*

	Modifications by Ladislav Kavan, 2002

*/

#include "stdafx.h"			// Header File For MFC
#include <gl\gl.h>			// Header File For The OpenGL32 Library

#include "MilkshapeModel.h"

#include <fstream.h>
#include <stdio.h>
#include <direct.h>

MilkshapeModel::MilkshapeModel()
{
}

MilkshapeModel::~MilkshapeModel()
{
	delete m_pTr;
	delete m_nonhitTr;
}

#include "MilkshapeStructures.h"

bool MilkshapeModel::loadModelData( const char *filename )
{
	ifstream inputFile( filename, ios::in | ios::binary | ios::nocreate );
	if ( inputFile.fail())
		return false;	// "Couldn't open the model file."

	char pathTemp[PATH_MAX+1];
	int pathLength;
	for ( pathLength = strlen( filename ); pathLength--; ) {
		if ( filename[pathLength] == '/' || filename[pathLength] == '\\' ) {
			break;
		}
	}
	strncpy( pathTemp, filename, pathLength );

	inputFile.seekg( 0, ios::end );
	long fileSize = inputFile.tellg();
	inputFile.seekg( 0, ios::beg );

	byte *pBuffer = new byte[fileSize];
	inputFile.read( pBuffer, fileSize );
	inputFile.close();

	const byte *pPtr = pBuffer;
	MS3DHeader *pHeader = ( MS3DHeader* )pPtr;
	pPtr += sizeof( MS3DHeader );

	if ( strncmp( pHeader->m_ID, "MS3D000000", 10 ) != 0 )
		return false; // "Not a valid Milkshape3D model file."

	if ( pHeader->m_version < 3 || pHeader->m_version > 4 )
		return false; // "Unhandled file version. Only Milkshape3D Version 1.3 and 1.4 is supported." );

	int nVertices = *( word* )pPtr; 
	m_numVertices = nVertices;
	m_origNumVertices = nVertices;
	m_pVertices = new Vertex[nVertices];
	m_pOrigVertices = new Vertex[nVertices];
	m_pActVers = new ActVertex[nVertices];
	for (int n=0; n<nVertices; n++) m_pActVers[n].m_location[0]=m_pActVers[n].m_location[1]=m_pActVers[n].m_location[2]=0.0;
	pPtr += sizeof( word );

	int i;
	for ( i = 0; i < nVertices; i++ )
	{
		MS3DVertex *pVertex = ( MS3DVertex* )pPtr;
		// v MS3D je vazba vrcholu k prave jedne kosti...
		m_pVertices[i].m_boneID = pVertex->m_boneID;

		m_pVertices[i].m_bone[0] = pVertex->m_boneID;
		m_pVertices[i].m_bone[1] = m_pVertices[i].m_bone[2] = m_pVertices[i].m_bone[3] = -1;
		m_pVertices[i].m_weight[0] = 1.0;
		m_pVertices[i].m_weight[1] = m_pVertices[i].m_weight[2] = m_pVertices[i].m_weight[3] = 0.0;
		memcpy( m_pVertices[i].m_location, pVertex->m_vertex, sizeof( float )*3 );

		memcpy( &m_pOrigVertices[i], &m_pVertices[i], sizeof( Vertex ));
		pPtr += sizeof( MS3DVertex );
	}

	int nTriangles = *( word* )pPtr;
	m_numTriangles = nTriangles;
	m_origNumTriangles = nTriangles;
	m_pTriangles = new Triangle[nTriangles];
	m_pOrigTriangles = new Triangle[nTriangles];
	pPtr += sizeof( word );

	for ( i = 0; i < nTriangles; i++ )
	{
		MS3DTriangle *pTriangle = ( MS3DTriangle* )pPtr;
		int vertexIndices[3] = { pTriangle->m_vertexIndices[0], pTriangle->m_vertexIndices[1], pTriangle->m_vertexIndices[2] };
		float t[3] = { 1.0f-pTriangle->m_t[0], 1.0f-pTriangle->m_t[1], 1.0f-pTriangle->m_t[2] };
		memcpy( m_pTriangles[i].m_vertexNormals, pTriangle->m_vertexNormals, sizeof( float )*3*3 );
		memcpy( m_pTriangles[i].m_s, pTriangle->m_s, sizeof( float )*3 );
		memcpy( m_pTriangles[i].m_t, t, sizeof( float )*3 );
		memcpy( m_pTriangles[i].m_vertexIndices, vertexIndices, sizeof( int )*3 );

		memcpy( &m_pOrigTriangles[i], &m_pTriangles[i], sizeof( Triangle ));
		pPtr += sizeof( MS3DTriangle );
	}

	int nGroups = *( word* )pPtr;
	m_numMeshes = nGroups;
	m_pMeshes = new Mesh[nGroups];
	pPtr += sizeof( word );
	for ( i = 0; i < nGroups; i++ )
	{
		pPtr += sizeof( byte );	// flags
		pPtr += 32;				// name

		word nTriangles = *( word* )pPtr;
		pPtr += sizeof( word );
		int *pTriangleIndices = new int[nTriangles];
		for ( int j = 0; j < nTriangles; j++ )
		{
			pTriangleIndices[j] = *( word* )pPtr;
			pPtr += sizeof( word );
		}

		char materialIndex = *( char* )pPtr;
		pPtr += sizeof( char );
	
		m_pMeshes[i].m_materialIndex = materialIndex;
		m_pMeshes[i].m_numTriangles = nTriangles;
		m_pMeshes[i].m_pTriangleIndices = pTriangleIndices;
	}

	int nMaterials = *( word* )pPtr;
	m_numMaterials = nMaterials;
	m_pMaterials = new Material[nMaterials];
	pPtr += sizeof( word );
	for ( i = 0; i < nMaterials; i++ )
	{
		MS3DMaterial *pMaterial = ( MS3DMaterial* )pPtr;
		memcpy( m_pMaterials[i].m_ambient, pMaterial->m_ambient, sizeof( float )*4 );
		memcpy( m_pMaterials[i].m_diffuse, pMaterial->m_diffuse, sizeof( float )*4 );
		memcpy( m_pMaterials[i].m_specular, pMaterial->m_specular, sizeof( float )*4 );
		memcpy( m_pMaterials[i].m_emissive, pMaterial->m_emissive, sizeof( float )*4 );
		m_pMaterials[i].m_shininess = pMaterial->m_shininess;
		if ( strncmp( pMaterial->m_texture, ".\\", 2 ) == 0 ) {
			// MS3D 1.5.x relative path
			strcpy( pathTemp + pathLength, pMaterial->m_texture + 1 );
			m_pMaterials[i].m_pTextureFilename = new char[strlen( pathTemp )+1];
			strcpy( m_pMaterials[i].m_pTextureFilename, pathTemp );
		}
		else {
			// MS3D 1.4.x or earlier - absolute path
			m_pMaterials[i].m_pTextureFilename = new char[strlen( pMaterial->m_texture )+1];
			strcpy( m_pMaterials[i].m_pTextureFilename, pMaterial->m_texture );
		}
		pPtr += sizeof( MS3DMaterial );
	}

//	reloadTextures(); -- zruseno kvuli vice kameram
	// Load Skeleton Definiton

//	float animFPS = *( float* )pPtr;
	pPtr += sizeof( float );

	// skip currentTime
	pPtr += sizeof( float );

//	int totalFrames = *( int* )pPtr;
	pPtr += sizeof( int );

//	m_totalTime = totalFrames*1000.0/animFPS;

	m_numJoints = *( word* )pPtr;
	pPtr += sizeof( word );

	m_pJoints = new Joint[m_numJoints];

	struct JointNameListRec
	{
		int m_jointIndex;
		const char *m_pName;
	};

	const byte *pTempPtr = pPtr;

	JointNameListRec *pNameList = new JointNameListRec[m_numJoints];
	m_swordholder = -1; m_midfinger = -1; m_headJoint = -1;
	m_ikj[0] = -1; m_ikj[1] = -1; m_ikj[2] = -1;
	for ( i = 0; i < m_numJoints; i++ )
	{
		MS3DJoint *pJoint = ( MS3DJoint* )pTempPtr;
		pTempPtr += sizeof( MS3DJoint );
		pTempPtr += sizeof( MS3DKeyframe )*( pJoint->m_numRotationKeyframes+pJoint->m_numTranslationKeyframes );

		pNameList[i].m_jointIndex = i;
		pNameList[i].m_pName = pJoint->m_name;

		if (strcmp(pJoint->m_name,"SwordHolder")==0) m_swordholder = i;
		else if (strcmp(pJoint->m_name,"RightHand")==0) m_ikj[HAND] = i;
		else if (strcmp(pJoint->m_name,"RightEllbow")==0) m_ikj[ELBOW] = i;
		else if (strcmp(pJoint->m_name,"RightShoulder")==0) m_ikj[SHOULDER] = i;
		else if (strcmp(pJoint->m_name,"RightFinger2")==0) m_midfinger = i;
		else if (strcmp(pJoint->m_name,"Head")==0) m_headJoint = i;
	}	

	for ( i = 0; i < m_numJoints; i++ )
	{
		MS3DJoint *pJoint = ( MS3DJoint* )pPtr;
		pPtr += sizeof( MS3DJoint );

		int j, parentIndex = -1;
		if ( strlen( pJoint->m_parentName ) > 0 )
		{
			for ( j = 0; j < m_numJoints; j++ )
			{
				if ( stricmp( pNameList[j].m_pName, pJoint->m_parentName ) == 0 )
				{
					parentIndex = pNameList[j].m_jointIndex;
					break;
				}
			}
			if ( parentIndex == -1 ) // Unable to find parent bone in MS3D file
				return false;
		}

		memcpy( m_pJoints[i].m_localRotation, pJoint->m_rotation, sizeof( float )*3 );
		memcpy( m_pJoints[i].m_localTranslation, pJoint->m_translation, sizeof( float )*3 );
		m_pJoints[i].m_parent = parentIndex;

		// preskocim keyframy
		pPtr += sizeof(MS3DKeyframe)*(pJoint->m_numTranslationKeyframes + 
			                          pJoint->m_numRotationKeyframes);
	}
	delete[] pNameList;

	setupJoints();

	delete[] pBuffer;   
	
	m_pTr = new byte[m_numTriangles];
	for (n=0; n<m_numTriangles; n++) m_pTr[n] = 0;
	m_nonhitTr = new bool[m_numTriangles];
	for (n=0; n<m_numTriangles; n++) m_nonhitTr[n] = false;

	return true;
}

bool MilkshapeModel::loadNormals(const char *filename)
{
 int n=0;
 Point3 *norm;
 norm = new Point3[m_numVertices];

 FILE *fin = fopen(filename, "rt");
 if (fin == NULL) return false;
 while (true)
 {
	fscanf(fin, "%f %f %f\n", &norm[n].x, &norm[n].y, &norm[n].z);
	if (feof(fin)) break;
	if (++n >= m_numVertices) 
	{
		TRACE0("Vic normal nez vrcholu!");
		return false;
	}
 }
 fclose(fin);

 for (n=0;n<m_numTriangles;n++)
 {
	 for (int k=0;k<3;k++)
	 {
		m_pTriangles[n].m_vertexNormals[k][0] = norm[m_pTriangles[n].m_vertexIndices[k]].x;
		m_pTriangles[n].m_vertexNormals[k][1] = norm[m_pTriangles[n].m_vertexIndices[k]].y;
		m_pTriangles[n].m_vertexNormals[k][2] = norm[m_pTriangles[n].m_vertexIndices[k]].z;
	 }
 }

 delete[] norm;
 return true;
}

bool MilkshapeModel::loadWeights(const char *filename)
{
 FILE *fin = fopen(filename, "rt");
 if (fin == NULL) return false;
 int v = 0;
 while (!feof(fin))  {
//	int puv = m_pVertices[v].m_bone[0];
//	TRACE("puvodne %i",m_pVertices[v].m_bone[0]);
	fscanf(fin,"%i %i %i %i %f %f %f %f\n",&m_pVertices[v].m_bone[0],&m_pVertices[v].m_bone[1],
		&m_pVertices[v].m_bone[2],&m_pVertices[v].m_bone[3],&m_pVertices[v].m_weight[0],
		&m_pVertices[v].m_weight[1],&m_pVertices[v].m_weight[2],&m_pVertices[v].m_weight[3]);
/*
	TRACE(" nove %i(%f) %i(%f) %i(%f) %i(%f)\n",m_pVertices[v].m_bone[0],m_pVertices[v].m_weight[0],m_pVertices[v].m_bone[1],m_pVertices[v].m_weight[1],
		  m_pVertices[v].m_bone[2],m_pVertices[v].m_weight[2],m_pVertices[v].m_bone[3],m_pVertices[v].m_weight[3]);

	bool fnd = false;
	for (int n=0; n<4; n++)  {
      if (puv == m_pVertices[v].m_bone[n]) fnd=true;
	}    
	if (fnd == false)  
	{
		m_pVertices[v].m_bone[0] = puv;
		m_pVertices[v].m_bone[1] = m_pVertices[v].m_bone[2] = m_pVertices[v].m_bone[3] = -1;
		m_pVertices[v].m_weight[0] = 1.0;
		m_pVertices[v].m_weight[1] = m_pVertices[v].m_weight[2] = m_pVertices[v].m_weight[3] = 0.0;
	}
	*/
/*
	int maxw = 0;
	for (n=0; n<4; n++) 
		if (m_pVertices[v].m_weight[n] > m_pVertices[v].m_weight[maxw]) maxw = n;

	int nk = m_pVertices[v].m_bone[maxw];

	{
		m_pVertices[v].m_bone[0] = nk;
		m_pVertices[v].m_bone[1] = m_pVertices[v].m_bone[2] = m_pVertices[v].m_bone[3] = -1;
		m_pVertices[v].m_weight[0] = 1.0;
		m_pVertices[v].m_weight[1] = m_pVertices[v].m_weight[2] = m_pVertices[v].m_weight[3] = 0.0;
	}
/**/

	v++;
 }
 if (v != m_numVertices) TRACE0("MilkshapeModel::loadWeights - nesedi pocet vrcholu\n");
 fclose(fin);
 return true;
}

void MilkshapeModel::saveModelData( const char *filename )
{
	FILE *f = fopen(filename, "wt");
	fprintf(f, "// MilkShape 3D ASCII\n\nFrames: 48\nFrame: 1\n\nMeshes: 1\n\"Mesh0\" 0 0\n");
	fprintf(f, "%i\n", m_numVertices);
	for (int n=0; n<m_numVertices; n++)
	{
		int fnd = -1;
		for (int t=0; t<m_numTriangles; t++)
		{
			if (m_pTriangles[t].m_vertexIndices[0] == n) { fnd = 0; break; }
			if (m_pTriangles[t].m_vertexIndices[1] == n) { fnd = 1; break; }
			if (m_pTriangles[t].m_vertexIndices[2] == n) { fnd = 2; break; }
		}
		if (fnd == -1) TRACE("podezrely: vrchol %i neni v trojuhelniku\n", n);
		fprintf(f,"0 %f %f %f %f %f %i\n", m_pVertices[n].m_location[0], 
			m_pVertices[n].m_location[1], m_pVertices[n].m_location[2], 
			m_pTriangles[t].m_s[fnd], m_pTriangles[t].m_t[fnd], m_pVertices[n].m_boneID);
	}
	fprintf(f, "%i\n", m_numVertices);
	for (n=0; n<m_numVertices; n++)
	{
		int fnd = -1;
		for (int t=0; t<m_numTriangles; t++)
		{
			if (m_pTriangles[t].m_vertexIndices[0] == n) { fnd = 0; break; }
			if (m_pTriangles[t].m_vertexIndices[1] == n) { fnd = 1; break; }
			if (m_pTriangles[t].m_vertexIndices[2] == n) { fnd = 2; break; }
		}
		if (fnd == -1) TRACE("podezrely: vrchol %i neni v trojuhelniku\n", n);
		fprintf(f, "%f %f %f\n", m_pTriangles[t].m_vertexNormals[fnd][0],
			m_pTriangles[t].m_vertexNormals[fnd][1], m_pTriangles[t].m_vertexNormals[fnd][2]);
	}
	fprintf(f, "%i\n", m_numTriangles);
	for (n=0; n<m_numTriangles; n++)
	{
		fprintf(f, "0 %i %i %i %i %i %i 1\n", m_pTriangles[n].m_vertexIndices[0],
			m_pTriangles[n].m_vertexIndices[1], m_pTriangles[n].m_vertexIndices[2],
			m_pTriangles[n].m_vertexIndices[0], m_pTriangles[n].m_vertexIndices[1],
			m_pTriangles[n].m_vertexIndices[2]);
	}
	fclose(f);
}

// saves mesh data in VolInt format, assuimg one mesh:
void MilkshapeModel::saveModelVolInt( const char *filename )
{
	FILE *f = fopen(filename, "wt");
	fprintf(f, "%i\n", m_numVertices);
	for (int n=0; n<m_numVertices; n++)
	{		
		fprintf(f,"%f %f %f\n", m_pVertices[n].m_location[0], 
			m_pVertices[n].m_location[1], m_pVertices[n].m_location[2]);
	}
	
	fprintf(f, "\n%i\n", m_numTriangles);
	for (n=0; n<m_numTriangles; n++)
	{
		fprintf(f, "3 %i %i %i\n", m_pTriangles[n].m_vertexIndices[0],
			m_pTriangles[n].m_vertexIndices[1], m_pTriangles[n].m_vertexIndices[2]);			
	}
	fclose(f);
}


/*  stara verze (staticky model):



#include "stdafx.h"			// Header File For MFC
#include <gl\gl.h>			// Header File For The OpenGL32 Library

#include "MilkshapeModel.h"

#include <fstream.h>

MilkshapeModel::MilkshapeModel()
{
}

MilkshapeModel::~MilkshapeModel()
{
}


// byte-align structures
#ifdef _MSC_VER
#	pragma pack( push, packing )
#	pragma pack( 1 )
#	define PACK_STRUCT
#elif defined( __GNUC__ )
#	define PACK_STRUCT	__attribute__((packed))
#else
#	error you must byte-align these structures with the appropriate compiler directives
#endif

typedef unsigned char byte;
typedef unsigned short word;

// File header
struct MS3DHeader
{
	char m_ID[10];
	int m_version;
} PACK_STRUCT;

// Vertex information
struct MS3DVertex
{
	byte m_flags;
	float m_vertex[3];
	char m_boneID;
	byte m_refCount;
} PACK_STRUCT;

// Triangle information
struct MS3DTriangle
{
	word m_flags;
	word m_vertexIndices[3];
	float m_vertexNormals[3][3];
	float m_s[3], m_t[3];
	byte m_smoothingGroup;
	byte m_groupIndex;
} PACK_STRUCT;

// Material information
struct MS3DMaterial
{
    char m_name[32];
    float m_ambient[4];
    float m_diffuse[4];
    float m_specular[4];
    float m_emissive[4];
    float m_shininess;	// 0.0f - 128.0f
    float m_transparency;	// 0.0f - 1.0f
    byte m_mode;	// 0, 1, 2 is unused now
    char m_texture[128];
    char m_alphamap[128];
} PACK_STRUCT;

//	Joint information
struct MS3DJoint
{
	byte m_flags;
	char m_name[32];
	char m_parentName[32];
	float m_rotation[3];
	float m_translation[3];
	word m_numRotationKeyframes;
	word m_numTranslationKeyframes;
} PACK_STRUCT;

// Keyframe data
struct MS3DKeyframe
{
	float m_time;
	float m_parameter[3];
} PACK_STRUCT;

// Default alignment
#ifdef _MSC_VER
#	pragma pack( pop, packing )
#endif

#undef PACK_STRUCT

bool MilkshapeModel::loadModelData( const char *filename )
{
	ifstream inputFile( filename, ios::in | ios::binary | ios::nocreate );
	if ( inputFile.fail())
		return false;	// "Couldn't open the model file."

	inputFile.seekg( 0, ios::end );
	long fileSize = inputFile.tellg();
	inputFile.seekg( 0, ios::beg );

	byte *pBuffer = new byte[fileSize];
	inputFile.read( pBuffer, fileSize );
	inputFile.close();

	const byte *pPtr = pBuffer;
	MS3DHeader *pHeader = ( MS3DHeader* )pPtr;
	pPtr += sizeof( MS3DHeader );

	if ( strncmp( pHeader->m_ID, "MS3D000000", 10 ) != 0 )
		return false; // "Not a valid Milkshape3D model file."

	if ( pHeader->m_version < 3 || pHeader->m_version > 4 )
		return false; // "Unhandled file version. Only Milkshape3D Version 1.3 and 1.4 is supported." );

	int nVertices = *( word* )pPtr; 
	m_numVertices = nVertices;
	m_pVertices = new Vertex[nVertices];
	pPtr += sizeof( word );

	int i;
	for ( i = 0; i < nVertices; i++ )
	{
		MS3DVertex *pVertex = ( MS3DVertex* )pPtr;
		m_pVertices[i].m_boneID = pVertex->m_boneID;
		memcpy( m_pVertices[i].m_location, pVertex->m_vertex, sizeof( float )*3 );
		pPtr += sizeof( MS3DVertex );
	}

	int nTriangles = *( word* )pPtr;
	m_numTriangles = nTriangles;
	m_pTriangles = new Triangle[nTriangles];
	pPtr += sizeof( word );

	for ( i = 0; i < nTriangles; i++ )
	{
		MS3DTriangle *pTriangle = ( MS3DTriangle* )pPtr;
		int vertexIndices[3] = { pTriangle->m_vertexIndices[0], pTriangle->m_vertexIndices[1], pTriangle->m_vertexIndices[2] };
		float t[3] = { 1.0f-pTriangle->m_t[0], 1.0f-pTriangle->m_t[1], 1.0f-pTriangle->m_t[2] };
		memcpy( m_pTriangles[i].m_vertexNormals, pTriangle->m_vertexNormals, sizeof( float )*3*3 );
		memcpy( m_pTriangles[i].m_s, pTriangle->m_s, sizeof( float )*3 );
		memcpy( m_pTriangles[i].m_t, t, sizeof( float )*3 );
		memcpy( m_pTriangles[i].m_vertexIndices, vertexIndices, sizeof( int )*3 );
		pPtr += sizeof( MS3DTriangle );
	}

	int nGroups = *( word* )pPtr;
	m_numMeshes = nGroups;
	m_pMeshes = new Mesh[nGroups];
	pPtr += sizeof( word );
	for ( i = 0; i < nGroups; i++ )
	{
		pPtr += sizeof( byte );	// flags
		pPtr += 32;				// name

		word nTriangles = *( word* )pPtr;
		pPtr += sizeof( word );
		int *pTriangleIndices = new int[nTriangles];
		for ( int j = 0; j < nTriangles; j++ )
		{
			pTriangleIndices[j] = *( word* )pPtr;
			pPtr += sizeof( word );
		}

		char materialIndex = *( char* )pPtr;
		pPtr += sizeof( char );
	
		m_pMeshes[i].m_materialIndex = materialIndex;
		m_pMeshes[i].m_numTriangles = nTriangles;
		m_pMeshes[i].m_pTriangleIndices = pTriangleIndices;
	}

	int nMaterials = *( word* )pPtr;
	m_numMaterials = nMaterials;
	m_pMaterials = new Material[nMaterials];
	pPtr += sizeof( word );
	for ( i = 0; i < nMaterials; i++ )
	{
		MS3DMaterial *pMaterial = ( MS3DMaterial* )pPtr;
		memcpy( m_pMaterials[i].m_ambient, pMaterial->m_ambient, sizeof( float )*4 );
		memcpy( m_pMaterials[i].m_diffuse, pMaterial->m_diffuse, sizeof( float )*4 );
		memcpy( m_pMaterials[i].m_specular, pMaterial->m_specular, sizeof( float )*4 );
		memcpy( m_pMaterials[i].m_emissive, pMaterial->m_emissive, sizeof( float )*4 );
		m_pMaterials[i].m_shininess = pMaterial->m_shininess;
		m_pMaterials[i].m_pTextureFilename = new char[strlen( pMaterial->m_texture )+1];
		strcpy( m_pMaterials[i].m_pTextureFilename, pMaterial->m_texture );
		pPtr += sizeof( MS3DMaterial );
	}

	reloadTextures();

	delete[] pBuffer;

	return true;
}

*/