/*
	Model.h

		Abstract base class for a model. The specific extended class will render the given model. 

	Author:	Brett Porter
	Email: brettporter@yahoo.com
	Website: http://www.geocities.com/brettporter/
	Copyright (C)2000, Brett Porter. All Rights Reserved.

	This file may be used only as long as this copyright notice remains intact.
*/

/*

	Modifications by Ladislav Kavan, 2002

	- removed skeletal animation (in separate class, independent of model)
	- added weapon model

*/

#ifndef MODEL_H
#define MODEL_H

#include "gl\gl.h"
#include "Matrix.h"
#include "Timer.h"
#include "constants.h"
#include <assert.h>

const int SHOULDER = 0;
const int ELBOW = 1;
const int HAND = 2;

const int BASIC_SA = 1;
const int VERTEX_BLENDING = 2;
const int BONES_BLENDING = 3;

// vyznacne vrcholy zbrane (kolik je treba ODECIST od celk. poctu vrcholu): (11.2.03)
const int vA = 6, vB = 5, vC = 4, vD = 3, vE = 2, vF = 1;

typedef struct SPoint3 
{
	float x, y, z;
} Point3;

class Model 
{
	friend class CDModel;
	friend class ColResp;
	public:		

		//	Mesh
		struct Mesh
		{
			int m_materialIndex;
			int m_numTriangles;
			int *m_pTriangleIndices;
		};

		//	Material properties
		struct Material
		{
			float m_ambient[4], m_diffuse[4], m_specular[4], m_emissive[4];
			float m_shininess;
			GLuint m_texture[cMaxCameras]; // pro kazdou kameru (rendering context) muze byt jine cislo textury
			char *m_pTextureFilename;
		};

		//	Triangle structure
		struct Triangle
		{
			float m_vertexNormals[3][3];
			float m_s[3], m_t[3];
			int m_vertexIndices[3];
		};

		//	Vertex structure
		struct Vertex
		{			
			char m_boneID; // kdybych mel vybrat jedinou kost, ke ktere by byl navazany
			char m_bone[4]; // ke kterym kostem navazany
			float m_weight[4]; // a s jakyma vahama
			float m_location[3];
		};

		struct ActVertex
		{
			float m_location[3];
			float m_normal[3];
		};
		
		//	Skeleton bone joint
		struct Joint
		{
			float m_localRotation[3];
			float m_localTranslation[3];
			Matrix m_absolute, m_relative; // potud vse z definice modelu

			// transformace v prostoru kloubu (preneseny do pocatku), ktere jsou vysledkem animace:
//			float m_jointRotation[3]; // okolo os X,Y,Z
//			float m_jointTranslation[3];
/*
			int m_numRotationKeyframes, m_numTranslationKeyframes;
			Keyframe *m_pTranslationKeyframes;
			Keyframe *m_pRotationKeyframes;

			int m_currentTranslationKeyframe, m_currentRotationKeyframe;
*/
			Matrix m_final; //spocita se z m_relative_otce*m_relative*(transofrmace dana m_jointRotation a m_jointTranslation)

			Matrix m_transform; //transformace kloubu po preneseni do pocatku	
			//float trans[3], rot[3]; //transformace kloubu po preneseni do pocatku

			int m_parent;
		};
		// { end NEW }

	public:
		/*	Constructor. */
		Model();

		/*	Destructor. */
		virtual ~Model();

		/*	
			Load the model data into the private variables. 
				filename			Model filename
		*/
		virtual bool loadModelData( const char *filename ) = 0;
		// Nastavi modelu obecnejsi prirazeni vrcholu kostem, nez uklada MS3D
		virtual bool loadWeights(const char *filename) = 0;
		virtual bool loadNormals(const char *filename) = 0;
		// Save model data in MS3D ASCII format
		virtual void saveModelData( const char *filename ) = 0;

		// pocita skutecne polohy vrcholu (podle m_final)
		void compActVers(bool weapon = false, Matrix *swordholder = 0);
		/*
			Draw the model - ale pomoci m_pActVers!
			camId - cislo kamery: potreba kvuli ruznym kodum textur pro ruzne RC
			weapon - drawing weapon or model?
			tr - pole trojuhelniku, ktere mam zvyraznit
			fade - kreslit "rozmazani" mece?
			transparency - pruhlednost modelu: 0 uplne pruhledny, 1 nepruhledny, -1 kresli normalne (s texturou)
		*/		
		void draw(int camId, bool weapon, bool fade = false, float transparency = -1);
		// na vsechny vrcholy aplikuje matici root_transform
		void transformVers(bool weapon = false);

		// natahne textury s kody pro kameru (RC) 'camId'
		void reloadTextures(int camId);

		int getNumJoints() { return m_numJoints; }
		int getNumTriangles() { return m_numTriangles; }
		int getNumVertices() { return m_numVertices; }
		Model *getWeapon() { return m_pWeapon; }
		ActVertex *getActVers() { return m_pActVers; }
		//Mesh *getLastMesh() { return m_pMeshes[m_numMeshes-1]; }
		const float *getSHF() { return m_pJoints[m_swordholder].m_final.getMatrix(); }
		void setSHF(const float *m) { m_pJoints[m_swordholder].m_final.set(m); }
		// nastavi v poli m_pTr zasazeny trojuhelnik cislo 'tri':
		void setTriangleHit(int tri, byte value = 255);
		// zjisti jak moc je trojuhelnik 'tri' zasazeny:		
		byte getTriangleHit(int tri);

	protected:

		/*	Setup joint matrices. */
		void setupJoints();	

		// vraci vsechny 3 sousedni trojuhelniky
		int sousTroj(int t, int s[3]);

	protected:
		//	Meshes used
		int m_numMeshes;
		Mesh *m_pMeshes;

		//	Materials used
		int m_numMaterials;
		Material *m_pMaterials;

		//	Triangles used
		int m_numTriangles;
		Triangle *m_pTriangles;

		// Original triangles (as loaded from file)
		int m_origNumTriangles;
		Triangle *m_pOrigTriangles;

		//	Vertices Used
		int m_numVertices;
		Vertex *m_pVertices;

		ActVertex *m_pActVers; // skutecne polohy vrcholu

		// Original vertices
		int m_origNumVertices;		
		Vertex *m_pOrigVertices;

		//	Joint information
		int m_numJoints;

        // dalsi model - zbran
		Model *m_pWeapon;

		// je levak?
		bool m_leftHanded; 

		// transformace modelu ktera se provede uplne pred vsemi ostatnimi
		// je to kvuli dvema sermirum (ruzne posunutym a otocenym)
		Matrix root_transform;

		byte *m_pTr; // pole zasazenych trojuhelniku

		bool *m_nonhitTr; // pole nezasahovych trojuhelniku (kam zasah neplati)

	public: 
		void healTri(byte sub);
		Joint *m_pJoints;// budou se menit z KeyframeInterpolatoru

		int m_swordholder; // index kloubu pro drzeni mece
		int m_midfinger; // index prostrednicku (pro urceni osy twistu zapesti)
		int m_headJoint; // index kloubu hlavy
		int m_ikj[3]; // inverse kinematics joints (SHOULDER, ELBOW, HAND)

		//bool m_BDdisabled; // disable bones deformation
		int	m_renderingMethod; // jedno z: BASIC_SA, VERTEX_BLENDING, BONES_BLENDING
		
		bool loadWeapon(const char *fname);
		bool getLeftHanded() { return m_leftHanded; }
		void setLeftHanded(bool s) { m_leftHanded = s; 
		  if (m_pWeapon) m_pWeapon -> setLeftHanded(s); // propagace do zbrane
		}

		// je trojuhelnik nezasahovy?
		bool isNonhit(int tr) { assert(tr>=0 && tr<m_numTriangles); return m_nonhitTr[tr]; }

		// nastavi nezasahove trojuhelniky:
		void setNonhitTriangles();

		// compute m_final in joints from current transformation
		void setFinalMatrices(const float *root_pos);		

		bool eqPoints(float p1[3], float p2[3])  {
			if ((fabs(p1[0] - p2[0]) < 0.001)&&(fabs(p1[1] - p2[1]) < 0.001)&&
				(fabs(p1[2] - p2[2]) < 0.001)) return true;
			return false;
		}
		bool eqVers(int v1, int v2)  {
			return eqPoints(m_pVertices[v1].m_location, m_pVertices[v2].m_location);
		}
		// compute the average curvature of triangles attached by all vertices to 'joint' (14.12.02):
		float getCurvature(int joint);

		// test duplicity vrcholu
		bool checkDup(void);

		// subdivide triangles attached to 'joint' (with all vertices)
		void subdivMesh1k3(int joint);		
		void subdivMesh1k4(int joint, int joint2 = -1);

		void setRootTransform(Matrix rt) { root_transform = rt; }
		const Matrix &getRootTransform() { return root_transform; }
	protected:
		int newVertex(Vertex newVer[], int &nv, Vertex &v1, Vertex &v2, int joint);
		void newTriangle(Triangle newTri[], int &nt, int v0, int v1, int v2,int t, int va0, int vb0, int va1, int vb1, int va2, int vb2);
		int protiVrchol(int v1, int v2, int vi[3]);
		// vraci true kdyz trojuhelnik nema byt deleny:
		bool podminka(const Vertex v[3], int joint, int joint2 = -1) {
			if (joint2 == -1)			
				return (v[0].m_boneID != joint || v[1].m_boneID != joint || v[2].m_boneID != joint);
			else
			{
				if (v[0].m_boneID == joint && v[1].m_boneID == joint2 && v[2].m_boneID == joint2) return false;
				if (v[0].m_boneID == joint2 && v[1].m_boneID == joint && v[2].m_boneID == joint2) return false;
				if (v[0].m_boneID == joint2 && v[1].m_boneID == joint2 && v[2].m_boneID == joint) return false;
				return true;
			}			
		}		
		// vraci prumer krivosti vsech tri hran trojuhelnika 't'
		float triangleCurv(int t);
		// vraci normalu 'N' trojuhelnika 't'
		void triangleNormal(int t, float N[3]);

};

#endif // ndef MODEL_H