/*
	Vector.h

		Vector manipulation.

	Author:	Brett Porter
	Email: brettporter@yahoo.com
	Website: http://rsn.gamedev.net/pl3d
	Copyright (C)2000, 2001, Brett Porter. All Rights Reserved.
	This source code is released under the LGPL. See license.txt for details.

	Created: 9 December 2000
	Last Edited: 8 January 2001

	Please see the file ChangeLog.html for a revision history.
*/

/*
	Modifications by Ladislav Kavan, 2002
*/

#ifndef PORTALIB3D_VECTOR_HXX
#define PORTALIB3D_VECTOR_HXX

#include "Matrix.h"

#pragma warning( disable : 4244 4305 )  

class Vector
{
	public:
		/*	Constructor. (0, 0, 0, 1) */
		Vector();

		/*	Constructor. 3 float values array. */
		Vector( const float *vector );

		/* Constructor. 3 float values. */
		Vector( float a, float b, float c);

		/*  Copy constructor */
		Vector( const Vector &v );

		/*	Destructor. */
		~Vector();

		/*	Retrieve vector. */
		const float *getVector() const { return m_vector; }

		/*	Transform the vector by a matrix. */
		void transform( const Matrix& m );

		/*	Transform the vector by a matrix, not including the scaling or transformation elements (use only top-left 3x3 of matrix). */
		void transform3( const Matrix& m );

		/*	Set the values of the vector. Takes 3 float values. */
		void set( const float *vector );

		/*	Translate by another vector. */
		void add( const Vector& v );

		/* Subtract another vector. */
		void sub( const Vector& v );

		/* Multiplication of coordinates */
		void mul( float a );

		/* Multiplication of coordinates */
		void mul( float a, float b, float c );

		/* Cross product */
		void cross( Vector &v );

		/* Get vector. 3 float values. */
		void get( float *v );

		/*	Reset to (0, 0, 0, 1). */
		void reset();

		/*	Get the length of the vector. */
		float length() const;

		/*	Normalize (make it a unit vector). */
		void normalize();

		/* Divide by last coordinate */
		void homog();

		/* Text output */
		void print();

		const float operator[]( int index ) const { return m_vector[index]; }

	private:
		//	Vector data
		float m_vector[4];
	public:

		static void add3(float *a, const float *b)  {
			a[0] += b[0]; a[1] += b[1]; a[2] += b[2];
		}

		static void sub3(float *a, const float *b)  {
			a[0] -= b[0]; a[1] -= b[1]; a[2] -= b[2];			
		}

		static void mul3(float *a, const float c)  {
			a[0]*=c; a[1]*=c; a[2]*=c;
		}

		static void set3(float *a, const float *b)  {
			a[0] = b[0]; a[1] = b[1]; a[2] = b[2];
		}

		static void norm3(float a[3]) {
			float len = sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);
			a[0]/=len; a[1]/=len; a[2]/=len;
		}
};

inline void Vector::reset()
{
	m_vector[0] = m_vector[1] = m_vector[2] = 0;
	m_vector[3] = 1;
}

inline void Vector::set( const float *values )
{
	m_vector[0] = values[0];
	m_vector[1] = values[1];
	m_vector[2] = values[2];
	m_vector[3] = 1.0;
}

inline void Vector::add( const Vector& v )
{
	m_vector[0] += v.m_vector[0];
	m_vector[1] += v.m_vector[1];
	m_vector[2] += v.m_vector[2];
//	m_vector[3] += v.m_vector[3];
}

inline void Vector::mul( float a, float b, float c )
{
	m_vector[0] *= a;
	m_vector[1] *= b;
	m_vector[2] *= c;
}

inline void Vector::mul( float a )
{
	m_vector[0] *= a;
	m_vector[1] *= a;
	m_vector[2] *= a;
}

inline void Vector::cross( Vector &v )
{
	homog();
	v.homog();
	float o[3];
	o[0] = m_vector[1]*v.m_vector[2] - v.m_vector[1]*m_vector[2];
	o[1] = m_vector[2]*v.m_vector[0] - v.m_vector[2]*m_vector[0];
	o[2] = m_vector[0]*v.m_vector[1] - v.m_vector[0]*m_vector[1];
	m_vector[0] = o[0]; m_vector[1] = o[1]; m_vector[2] = o[2];
}

inline void Vector::get( float *v )
{
	v[0] = m_vector[0] / m_vector[3];
	v[1] = m_vector[1] / m_vector[3];
	v[2] = m_vector[2] / m_vector[3];
}

inline void Vector::homog()
{
	m_vector[0] = m_vector[0] / m_vector[3];
	m_vector[1] = m_vector[1] / m_vector[3];
	m_vector[2] = m_vector[2] / m_vector[3];
	m_vector[3] = 1.0;
}

inline void Vector::sub( const Vector& v )
{
	homog();
	Vector v2(v);
	v2.homog();
	m_vector[0] -= v2.m_vector[0];
	m_vector[1] -= v2.m_vector[1];
	m_vector[2] -= v2.m_vector[2];
	m_vector[3] = 1.0;
}

inline void Vector::normalize()
{
	float len = length();

	m_vector[0] /= len;
	m_vector[1] /= len;
	m_vector[2] /= len;
}

inline float Vector::length() const
{
	return ( float )sqrt( m_vector[0]*m_vector[0]+m_vector[1]*m_vector[1]+m_vector[2]*m_vector[2] );
}

inline void Vector::print()
{
	homog();
	TRACE("%f %f %f\n",m_vector[0],m_vector[1],m_vector[2]);
}

#endif // ndef PORTALIB3D_VECTOR_HXX
