/* 
    geometry.cpp

	Auxiliary geometric routines

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include <assert.h>
#include "myvec.h"
#include "Vector.h"

// odsoupne bod B0 od bodu A0 ve smeru s na vzdalenost D; vraci o kolik je treba
// posunout B0 ve smeru s (kladne!) (12.2.02); neexistuje-li reseni, vraci -1
float sepPoints(const float A0[3], const float B0[3], const float s[3], float D)
{
	float M, N;
	float sKvad = DOT(s, s);
	float tmp[3];
	cpvector(tmp, A0);
	Vector::sub3(tmp, B0);
	N = (DOT(tmp, tmp) - D*D) / sKvad;
	M = -2 * DOT(tmp, s) / sKvad;	
	float disk = M*M - 4*N;
	if (disk < 0) return -1;
	float p1 = (-M + sqrt(disk)) / 2;
	float p2 = (-M - sqrt(disk)) / 2;
	float pmax;
	if (p2>p1) pmax = p2;
	else pmax = p1;
	if (pmax < 0) return -1;
	return pmax;
}

// res = (1-t)*A + t*B
void linterpvect(float res[3], const float A[3], const float B[3], float t)
{
	float tmp[3];
	cpvector(tmp, A);
	Vector::mul3(tmp, 1-t);
	cpvector(res, B);
	Vector::mul3(res, t);
	Vector::add3(res, tmp);
}

// viz. poznamky 6.9.02

// vraci vzdalenost mezi body A a B
float distPointPoint(const float A[3], const float B[3])
{
	float d[3];
	cpvector(d, A);
	Vector::sub3(d, B);
	return norm(d);
}

// vraci vzdalenost mezi useckou danou body M, P a bodem N
// a nastavuje bod na usecce B, kde se minima nabyva
float distSegPoint(const float M[3], const float P[3], const float N[3], float B[3])
{
	float w[3],NbM[3];
	cpvector(w, P);
	Vector::sub3(w, M);
	cpvector(NbM, N);
	Vector::sub3(NbM, M);
	float normw2 = DOT(w,w);
	assert(normw2>0);
	float r = DOT(NbM, w)/normw2;
	if (r>=0 && r<=1)
	{
		cpvector(B, w);
		Vector::mul3(B, r);
		Vector::add3(B, M);
		return distPointPoint(B, N);
	}
	else
	{ // minima se nabylo mimo usecku, musime prozkoumat krajni body
		float v1 = distPointPoint(N, M);
		float v2 = distPointPoint(N, P);
		if (v1<v2)
		{
			cpvector(B, M);
			return v1;
		}
		else
		{
			cpvector(B, P);
			return v2;
		}
	}
}


// vraci vzdalenost mezi useckou danou body A, B a
// useckou danou body C, D; nastavuje body X, Y kde lezi minimalni pricka
// vraci -1 jsou-li usecky rovnobezne
float distSegSeg(const float A[3], const float B[3], const float C[3], const float D[3], 
				 float X[3], float Y[3])
{
	float u[3], v[3], un[3], vn[3], AbC[3];
	cpvector(u, B);Vector::sub3(u, A);
	cpvector(v, D);Vector::sub3(v, C);
	cpvector(un, u); float ulen = unitize(un);
	cpvector(vn, v); float vlen = unitize(vn);
	cpvector(AbC, A); Vector::sub3(AbC, C);
	float K0 = DOT(un, vn), K1 = DOT(AbC, un), K2 = DOT(AbC, vn);
	if (fabs(K0*K0-1) < 0.00001) 
	{
		TRACE("singularita: rovnobezne usecky\n");
		// minima se nabyva v jednom z bodu, urci se pozdeji (stejne jako pro minimum mimo ruznobezy)
	}
	else
	{ // ruznobezky:
		float s = (K1*K0-K2)/(K0*K0-1);
		float t = (K1-K2*K0)/(K0*K0-1);
		if (s>=0 && s<=vlen && t>=0 && t<=ulen)
		{
			cpvector(X, un);
			Vector::mul3(X, t);
			Vector::add3(X, A);
			cpvector(Y, vn);
			Vector::mul3(Y, s);
			Vector::add3(Y, C);
			return distPointPoint(X, Y);
		}
	}
	// bud rovnobezky, nebo minimum mimo usecky:
	// musim zkoumat vzdalenosti bodu s useckama
	float bod[4][3];
	float vzd[4];

	vzd[0] = distSegPoint(C, D, A, bod[0]); // A s useckou CD
	vzd[1] = distSegPoint(C, D, B, bod[1]); // B s useckou CD
	vzd[2] = distSegPoint(A, B, C, bod[2]); // C s useckou AB
	vzd[3] = distSegPoint(A, B, D, bod[3]); // D s useckou AB

	int minidx = 0;
	for (int n=1; n<4; n++) if (vzd[n] < vzd[minidx]) minidx = n;		
	
	if (minidx==0) { cpvector(X, A); cpvector(Y, bod[0]); }
	else if (minidx==1) { cpvector(X, B); cpvector(Y, bod[1]); }
	else if (minidx==2) { cpvector(X, bod[2]); cpvector(Y, C); }
	else if (minidx==3) { cpvector(X, bod[3]); cpvector(Y, D); }

	return vzd[minidx];
}


void geometry_test()
{
	/* testik sepPoints:
	float A[3] = {0, 0, 0};
	float B[3] = {1, 1, 0};
	float s[3] = {0, -2, 0};

	float res = sepPoints(A, B, s, 2);
	TRACE("geometry_test: res = %f\n", res);
	float Bpos[3];
	cpvector(Bpos, s);
	Vector::mul3(Bpos, res);
	Vector::add3(Bpos, B);
	float dist = distPointPoint(A, Bpos);
	TRACE("geometry_test: dist = %f\n", dist);
	/**/

	/*
	float A[3] = {-2, -3, 2};
	float B[3] = {2, -1, 1};
	float C[3] = {1, 6, 2};
	float D[3] = {1, 7, 1};
	float X[3], Y[3];
	float dist = distSegSeg(A, B, C, D, X, Y);

	int n=1;	
	*/
}